from deprecated import deprecated
from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import (
    DICT_41_AUTO_8_SPEEDS,
    DICT_41_ON_OFF,
    DICT_41_YES_NO,
    _int,
    _swap_dict,
)

FAN_SPEED = _swap_dict(DICT_41_AUTO_8_SPEEDS)

PHOTOCATALYST_STATUS = _swap_dict(DICT_41_YES_NO)


# ----- Air Cleaner Class -------
# filter change status notify
@deprecated(reason="Scheduled for removal.")
def _0135E1(edt):
    return _int(edt, DICT_41_YES_NO)


# air volume 0x41 for auto
@deprecated(reason="Scheduled for removal.")
def _0135A0(edt):
    return _int(edt, DICT_41_AUTO_8_SPEEDS)


# cigarette sensor status
@deprecated(reason="Scheduled for removal.")
def _0135C1(edt):
    return _int(edt, DICT_41_ON_OFF)


# Photocatalyst setting
@deprecated(reason="Scheduled for removal.")
def _0135C2(edt):
    return _int(edt, DICT_41_ON_OFF)


# air pollution status
@deprecated(reason="Scheduled for removal.")
def _0135C0(edt):
    return _int(edt, {0x41: "pollution", 0x42: "fresh"})


"""Class for Air Cleaner Objects"""


class HomeAirCleaner(EchonetInstance):
    EPC_FUNCTIONS = {
        0xE1: [_int, DICT_41_YES_NO],
        0xA0: [_int, DICT_41_AUTO_8_SPEEDS],
        0xC0: [
            _int,
            {
                0x41: "pollution",
                0x42: "fresh",
            },
        ],
        0xC1: [_int, DICT_41_YES_NO],
        0xC2: [_int, DICT_41_ON_OFF],
    }

    def __init__(self, host, api_connector=None, instance=0x1):
        self._eojgc = 0x01  # Air conditioner-related device group
        self._eojcc = 0x35  # Air Cleaner
        EchonetInstance.__init__(
            self, host, self._eojgc, self._eojcc, instance, api_connector
        )

    """
    getFilterChangeStatus get the status of filter status.

    return: A string representing the lock status
    """

    def getFilterChangeStatus(self):
        return self.getMessage(0xE1)

    """
    setFanSpeed set the desired fan speed (e.g Low, Medium, High etc)

    param fans_speed: A string representing the fan speed
    """

    def setFanSpeed(self, fan_speed):
        return self.setMessage(0xA0, FAN_SPEED[fan_speed])

    """
    GetFanSpeed gets the current fan speed (e.g Low, Medium, High etc)
    Refer EPC code 0xA0: ('Air flow rate setting')

    return: A string representing the fan speed
    """

    def getFanSpeed(self):  # 0xA0
        return self.getMessage(0xA0)

    """
    getCigaretteSensorStatus get the status of cigarette sensor status.

    return: A string representing the lock status
    """

    def getCigaretteSensorStatus(self):
        return self.getMessage(0xC1)

    """
    getAirPollutionStatus get the status of air pollution status.

    return: A string representing the lock status
    """

    def getAirPollutionStatus(self):
        return self.getMessage(0xC0)

    """
    getPhotocatalystStatus get the status of Photocatalyst status.

    return: A string representing the lock status
    """

    def getPhotocatalystStatus(self):
        return self.getMessage(0xC2)

    """
    setPhotocatalyst set the Photocatalyst status

    param photocatalyst_status: A string representing the Photocatalyst speed
    """

    def setPhotocatalyst(self, photocatalyst_status):
        return self.setMessage(0xC2, PHOTOCATALYST_STATUS[photocatalyst_status])



from deprecated import deprecated
from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import (
    DICT_30_ON_OFF,
    DICT_41_HEATING_NOT_HEATING,
    DICT_41_ON_OFF,
    _int,
    _hh_mm,
)

# ----- Hot Water Generator -------


@deprecated(reason="Scheduled for removal.")
def _0272D0E2(edt):
    return _int(edt, DICT_41_HEATING_NOT_HEATING)


@deprecated(reason="Scheduled for removal.")
def _0272D2(edt):
    return _int(
        edt,
        {
            0x41: "Hot water warmer operation",
            0x42: "Hot water warmer operation resetting",
        },
    )


@deprecated(reason="Scheduled for removal.")
def _0272EX(edt):
    return _int(edt, DICT_41_ON_OFF)


@deprecated(reason="Scheduled for removal.")
def _0272D7(edt):
    return _int(edt, DICT_30_ON_OFF)


@deprecated(reason="Scheduled for removal.")
def _0272E8(edt):
    return _int(
        edt,
        {
            0x31: "Level 1",
            0x32: "Level 2",
            0x33: "Level 3",
            0x34: "Level 4",
            0x35: "Level 5",
            0x36: "Level 6",
            0x37: "Level 7",
            0x38: "Level 8",
        },
    )


def _0272DA(edt):  # Duration of automatic operation setting
    if edt == 0xFFFF:
        return "Limitless"
    hh = int.from_bytes(edt[0:1], "big")
    mm = str(int.from_bytes(edt[1:2], "big")).zfill(2)
    return f"{hh}:{mm}"


def _0272DB(edt):  # Remaining automatic operation time
    if edt == 0xFFFF:
        return "Infinite"
    hh = int.from_bytes(edt[0:1], "big")
    mm = str(int.from_bytes(edt[1:2], "big")).zfill(2)
    return f"{hh}:{mm}"


@deprecated(reason="Scheduled for removal.")
def _0272EF(edt):  # "Bath operation status monitor",
    return _int(
        edt,
        {
            0x41: "Supplying Hot Water",
            0x43: "Keeping Bath Temperature",
            0x42: "Stopped",
        },
    )


class HotWaterGenerator(EchonetInstance):
    EPC_FUNCTIONS = {
        0xD0: [
            _int,
            DICT_41_HEATING_NOT_HEATING,
        ],  # Hot water heating status
        0xD1: _int,  # Set value of hot water temperature
        0xD2: [
            _int,
            {
                0x41: "Hot water warmer operation",
                0x42: "Hot water warmer operation resetting",
            },
        ],  # Hot water warmer setting
        0xDA: _0272DA,  # Duration of automatic operation setting
        0xDB: _0272DB,  # Remaining automatic operation time
        0xE1: _int,  # Set value of bath of the bath temperature in degrees C
        0xE2: [
            _int,
            DICT_41_HEATING_NOT_HEATING,
        ],  # Bath water heater status
        0xE3: [_int, DICT_41_ON_OFF],  # Bath automatic mode setting
        0xE4: [_int, DICT_41_ON_OFF],  # Bath additional boil-up operation setting
        0xE5: [_int, DICT_41_ON_OFF],  # Bath adding hot water operation setting
        0xE6: [
            _int,
            DICT_41_ON_OFF,
        ],  # Bath water temperature lowering operation setting
        0xE7: _int,  # Bath hot water volume setting 1
        0xE8: [
            _int,
            {
                0x31: "Level 1",
                0x32: "Level 2",
                0x33: "Level 3",
                0x34: "Level 4",
                0x35: "Level 5",
                0x36: "Level 6",
                0x37: "Level 7",
                0x38: "Level 8",
            },
        ],  # Bath hot water volume setting 2
        0xEE: _int,  # Bath hot water volume setting 3
        0xD4: _int,  # Bath hot water volume setting 4
        0xD5: _int,  # Bath hot water volume setting 4 - Maximum settable level
        0xE9: [_int, DICT_41_ON_OFF],  # "Bathroom priority setting",
        0xEA: [_int, DICT_41_ON_OFF],  # "Shower hot water supply status",
        0xEB: [_int, DICT_41_ON_OFF],  # "Kitchen hot water supply status",
        0xEC: [
            _int,
            DICT_41_ON_OFF,
        ],  # "Hot water warmer ON timer reservation setting",
        0xED: _hh_mm,  # "Set value of hot water warmer ON timer time",
        0xEF: [
            _int,
            {
                0x41: "Supplying Hot Water",
                0x43: "Keeping Bath Temperature",
                0x42: "Stopped",
            },
        ],  # "Bath operation status monitor",
        0x90: [_int, DICT_41_ON_OFF],  # "ON timer reservation setting",
        0x91: _hh_mm,  # "Set value of ON timer time",
        0x92: _hh_mm,  # "Set value of ON timer relative time",
        0xD6: _int,  # "Volume setting",
        0xD7: [_int, DICT_30_ON_OFF],  # "Mute setting",
    }

    def __init__(self, host, api_connector=None, instance=0x1):
        self._eojgc = 0x02
        self._eojcc = 0x72
        EchonetInstance.__init__(
            self, host, self._eojgc, self._eojcc, instance, api_connector
        )



from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import _int

ENL_SCENE = 0xC0
ENL_SCENE_MAX = 0xC1

class LightingSystem(EchonetInstance):
    EPC_FUNCTIONS = {
        0xB0: _int,  # Illuminance level setting
        0xC0: _int,  # Scene control setting
        0xC1: _int,  # Max scene control setting
    }

    def __init__(self, host, api_connector=None, instance=0x1):
        self._eojgc = 0x02  # Housing/facility-related device group
        self._eojcc = 0xA3  # Lighting System class
        EchonetInstance.__init__(
            self, host, self._eojgc, self._eojcc, instance, api_connector
        )



import asyncio
from types import _ReturnT_co
from deprecated import deprecated
from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import (
    DATA_STATE_OFF,
    DATA_STATE_ON,
    DICT_41_AUTO_8_SPEEDS,
    DICT_41_AUTO_NONAUTO,
    DICT_41_ON_OFF,
    _int,
    _hh_mm,
    _signed_int,
)


def _014290(data):
    # Logic to interpret data for EPC 0x90
    return _int(data, DICT_41_AUTO_NONAUTO)


def _0142A0(data):
    # Logic to interpret data for EPC 0xA0
    return _int(data, DICT_41_AUTO_8_SPEEDS)


@deprecated(reason="Scheduled for removal.")
def _0142B1(data):
    # Logic to interpret data for EPC 0xB1
    return _int(data, DICT_41_ON_OFF)


@deprecated(reason="Scheduled for removal.")
def _014294(data):
    # Logic to interpret data for EPC 0xB1
    return _int(data, DICT_41_ON_OFF)


class ElectricHeater(EchonetInstance):
    EOJGC = 0x01
    EOJCC = 0x42

    EPC_FUNCTIONS = {
        0x80: _int,
        0xB1: [_int, DICT_41_AUTO_NONAUTO],
        0xB3: _int,
        0xBB: _signed_int,
        0xBC: _int,
        0xA0: [_int, DICT_41_AUTO_8_SPEEDS],
        0x90: [_int, DICT_41_ON_OFF],
        0x91: _hh_mm,
        0x92: _hh_mm,
        0x94: [_int, DICT_41_ON_OFF],
        0x95: _hh_mm,
        0x96: _hh_mm,
    }

    def __init__(self, host, api_connector=None, instance=0x1):
        self._eojgc = self.EOJGC
        self._eojcc = self.EOJCC
        super().__init__(host, self._eojgc, self._eojcc, instance, api_connector)

    @property
    async def operation_status(self):
        return await self.getMessage(0x80)

    @operation_status.setter
    async def operation_status(self, value):
        await self.setMessage(0x80, value)

    @property
    async def automatic_temperature_control_setting(self):
        return await self.getMessage(0xB1)

    @automatic_temperature_control_setting.setter
    async def automatic_temperature_control_setting(self, value):
        await self.setMessage(0xB1, value)

    @property
    async def temperature_setting(self):
        return await self.getMessage(0xB3)

    @temperature_setting.setter
    async def temperature_setting(self, value):
        await self.setMessage(0xB3, value)

    @property
    async def measured_room_temperature(self):
        return await self.getMessage(0xBB)

    @measured_room_temperature.setter
    async def measured_room_temperature(self, value):
        await self.setMessage(0xBB, value)

    @property
    async def remotely_set_temperature(self):
        return await self.getMessage(0xBC)

    @remotely_set_temperature.setter
    async def remotely_set_temperature(self, value):
        await self.setMessage(0xBC, value)

    @property
    async def air_flow_rate_setting(self):
        return await self.getMessage(0xA0)

    @air_flow_rate_setting.setter
    async def air_flow_rate_setting(self, value):
        await self.setMessage(0xA0, value)



from deprecated import deprecated
from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import _int, _signed_int, _to_string
import struct


@deprecated(reason="Scheduled for removal.")
def _0281D0(edt):
    # Water flow meter classification
    return _int(
        edt,
        {
            0x30: "Running Water",
            0x31: "Recycled Water",
            0x32: "Sewage Water",
            0x33: "Other Water",
        },
    )


@deprecated(reason="Scheduled for removal.")
def _0281D1(edt):
    # Owner classification
    return _int(
        edt,
        {
            0x30: "Not specified",
            0x31: "Public waterworks company",
            0x32: "Private sector company",
            0x33: "Individual",
        },
    )


@deprecated(reason="Scheduled for removal.")
def _0281E1(edt):
    # Unit for measured cumulative amounts of flowing water
    return _int(
        edt,
        {
            0x00: 1,
            0x01: 0.1,
            0x02: 0.01,
            0x03: 0.001,
            0x04: 0.0001,
            0x05: 0.00001,
            0x06: 0.000001,
        },
        None,
    )


# 0xE2: "Historical data of measured cumulative amounts of flowing water"
def _0281E2(edt):
    # return array x 48 unsigned long big-endian
    return [x[0] for x in struct.iter_unpack(">L", edt)]


class WaterFlowMeter(EchonetInstance):
    EPC_FUNCTIONS = {
        0xD0: [
            _int,
            {
                0x30: "Running Water",
                0x31: "Recycled Water",
                0x32: "Sewage Water",
                0x33: "Other Water",
            },
        ],  # "Water flow meter classification"
        0xD1: [
            _int,
            {
                0x30: "Not specified",
                0x31: "Public waterworks company",
                0x32: "Private sector company",
                0x33: "Individual",
            },
        ],  # "Owner classification"
        0xE0: _int,  # "Measured  cumulative amount of flowing water"
        0xE1: [
            _int,
            {
                0x00: 1,
                0x01: 0.1,
                0x02: 0.01,
                0x03: 0.001,
                0x04: 0.0001,
                0x05: 0.00001,
                0x06: 0.000001,
            },
            None,
        ],  # "Unit for measured cumulative amounts of flowing water"
        0xE2: _0281E2,  # "Historical data of measured cumulative amounts of flowing water"
        0xE5: _to_string,  # "ID number setting"
        0xE6: _to_string,  # "Verification expiration information"
    }

    def __init__(self, host, api_connector=None, instance=0x1):
        self._eojgc = 0x02
        self._eojcc = 0x81
        EchonetInstance.__init__(
            self, host, self._eojgc, self._eojcc, instance, api_connector
        )



from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import _int, _signed_int

ENL_DPM_ENG_NOR = 0xC0
ENL_DPM_ENG_REV = 0xC1
ENL_DPM_ENG_UNIT = 0xC2
ENL_DPM_DAY_GET_HISTORY = 0xC5
ENL_DPM_INSTANT_ENG = 0xC6
ENL_DPM_INSTANT_CUR = 0xC7
ENL_DPM_INSTANT_VOL = 0xC8
ENL_DPM_CHANNEL_SIMPLEX_CUMULATIVE_ENG = 0xB3
ENL_DPM_CHANNEL_SIMPLEX_INSTANT_ENG = 0xB7


def _0287C2(edt):
    op_mode = int.from_bytes(edt, "big")
    values = {
        0x00: 1,
        0x01: 0.1,
        0x02: 0.01,
        0x03: 0.001,
        0x04: 0.0001,
        0x0A: 10,
        0x0B: 100,
        0x0C: 1000,
        0x0D: 10000,
    }
    return values.get(op_mode, None)


def _0287B3(edt):
    values = []
    for x in range(0, edt[1]):
        values.append(
            int.from_bytes(edt[x * 4 + 2 : (x + 1) * 4 + 2], "big", signed=False)
        )
    return {"start": edt[0], "range": edt[1], "values": values}


def _0287B7(edt):
    values = []
    for x in range(0, edt[1]):
        values.append(
            int.from_bytes(edt[x * 4 + 2 : (x + 1) * 4 + 2], "big", signed=True)
        )
    return {"start": edt[0], "range": edt[1], "values": values}


# def _0287xx(edt):
#     return "Not implemented"
#
# def _0287D0EF(edt):
#     return "Measurement channels not implemented"


def _0287C7(edt):
    r_phase = float(int.from_bytes(edt[0:2], "big", signed=True)) / 10  # R Phase
    t_phase = float(int.from_bytes(edt[2:4], "big", signed=True)) / 10  # T Phase
    if t_phase == 3276.6:
        t_phase = None
    return {"r_phase_amperes": r_phase, "t_phase_amperes": t_phase}


def _0287C8(edt):
    r_sn = float(int.from_bytes(edt[0:2], "big", signed=True)) / 10  # R Phase
    sn_t = float(int.from_bytes(edt[2:4], "big", signed=True)) / 10  # T Phase
    if sn_t == 6553.4:
        sn_t = None
    return {"r_sn_voltage": r_sn, "sn_t_voltage": sn_t}


class DistributionPanelMeter(EchonetInstance):
    EPC_FUNCTIONS = {
        0xC0: _int,  # "Measured cumulative amount of electric energy (normal direction)"
        0xC1: _int,  # "Measured cumulative amount of electric energy (reverse direction)"
        0xC2: _0287C2,  # "Unit for cumulative amounts of electric energy"
        #       0xC3: _0287xx,     # "Historical data of measured cumulative amounts of electric energy (normal direction)"
        #       0xC4: _0287xx,     # "Historical data of measured cumulative amounts of electric energy (reverse direction)"
        0xC5: _int,  # "Day for which the historical data of measured cumulative amounts of electric energy is to be retrieved"
        0xC6: _signed_int,  # "Measured instantaneous amount of electric energy"
        0xC7: _0287C7,  # "Measured instantaneous currents"
        0xC8: _0287C8,  # "Measured instantaneous voltages"
        #       0xD0: _0287D0EF,   # "Measurement channel 1"
        #       0xD1: _0287D0EF,   # "Measurement channel 2"
        #       0xD2: _0287D0EF,   # "Measurement channel 3"
        #       0xD3: _0287D0EF,   # "Measurement channel 4"
        #       0xD4: _0287D0EF,   # "Measurement channel 5"
        #       0xD5: _0287D0EF,   # "Measurement channel 6"
        #       0xD6: _0287D0EF,   # "Measurement channel 7"
        #       0xD7: _0287D0EF,   # "Measurement channel 8"
        #       0xD8: _0287D0EF,   # "Measurement channel 9"
        #       0xD9: _0287D0EF,   # "Measurement channel 10"
        #       0xDA: _0287D0EF,   # "Measurement channel 11"
        #       0xDB: _0287D0EF,   # "Measurement channel 12"
        #       0xDC: _0287D0EF,   # "Measurement channel 13"
        #       0xDD: _0287D0EF,   # "Measurement channel 14"
        #       0xDE: _0287D0EF,   # "Measurement channel 15"
        #       0xDF: _0287D0EF,   # "Measurement channel 16"
        #       0xE0: _0287D0EF,   # "Measurement channel 17"
        #       0xE1: _0287D0EF,   # "Measurement channel 18"
        #       0xE2: _0287D0EF,   # "Measurement channel 19"
        #       0xE3: _0287D0EF,   # "Measurement channel 20"
        #       0xE4: _0287D0EF,   # "Measurement channel 21"
        #       0xE5: _0287D0EF,   # "Measurement channel 22"
        #       0xE6: _0287D0EF,   # "Measurement channel 23"
        #       0xE7: _0287D0EF,   # "Measurement channel 24"
        #       0xE8: _0287D0EF,   # "Measurement channel 25"
        #       0xE9: _0287D0EF,   # "Measurement channel 26"
        #       0xEA: _0287D0EF,   # "Measurement channel 27"
        #       0xEB: _0287D0EF,   # "Measurement channel 28"
        #       0xEC: _0287D0EF,   # "Measurement channel 29"
        #       0xED: _0287D0EF,   # "Measurement channel 30"
        #       0xEE: _0287D0EF,   # "Measurement channel 31"
        #       0xEF: _0287D0EF,   # "Measurement channel 32"
        0xB0: _int,  # "Master rated capacity"
        0xB1: _int,  # "Number of measurement channels (simplex)"
        #       0xB2: _0287xx,     # "Channel range specification for cumulative amount of electric power consumption measurement (simplex)"
        0xB3: _0287B3,  # "Measured cumulative amount of electric power consumption list (simplex)"
        #       0xB4: _0287xx,     # "Channel range specification for instantaneous current measurement (simplex)"
        #       0xB5: _0287xx,     # "Measured instantaneous current list (simplex)"
        #       0xB6: _0287xx,     # "Channel range specification for instantaneous power consumption measurement (simplex)"
        0xB7: _0287B7,  # "Measured instantaneous power consumption list (simplex)"
        0xB8: _int,  # "Number of measurement channels (duplex)"
        #       0xB9: _0287xx,     # "Channel range specification for cumulative amount of electric power consumption measurement (duplex)"
        #       0xBA: _0287xx,     # "Measured cumulative amount of electric power consumption list (duplex)"
        #       0xBB: _0287xx,     # "Channel range specification for instantaneous current measurement (duplex)"
        #       0xBC: _0287xx,     # "Measured instantaneous current list (duplex)"
        #       0xBD: _0287xx,     # "Channel range specification for instantaneous power consumption measurement (duplex)"
        #       0xBE: _0287xx,     # "Measured instantaneous power consumption list (duplex)"
    }

    def __init__(self, host, api_connector, instance=0x1):
        self._eojgc = 0x02
        self._eojcc = 0x87
        EchonetInstance.__init__(
            self, host, self._eojgc, self._eojcc, instance, api_connector
        )



from pychonet import EchonetInstance
from pychonet.lib.epc_functions import (
    DATA_STATE_OFF,
    DATA_STATE_ON,
    DICT_41_AUTO_NONAUTO,
    DICT_41_ON_OFF,
    _hh_mm,
    _int,
    _signed_int,
)


class FanHeater(EchonetInstance):
    class_codes = {"class_group_code": 0x01, "class_code": 0x43}

    EPC_FUNCTIONS = {
        0xB3: _int,
        0xBB: _signed_int,
        0xB1: [_int, DICT_41_AUTO_NONAUTO],
        0x90: [
            _int,
            {
                0x41: DATA_STATE_ON,
                0x42: DATA_STATE_OFF,
                0x43: "Timer-based ON",
                0x44: "Relative time ON",
            },
        ],
        0x91: _hh_mm,
        0x92: _hh_mm,
        0x94: [
            _int,
            {
                0x41: DATA_STATE_ON,
                0x42: DATA_STATE_OFF,
                0x43: "Timer-based ON",
                0x44: "Relative time ON",
            },
        ],
        0x95: _hh_mm,
        0x96: _hh_mm,
        0xC0: [_int, DICT_41_ON_OFF],
        0xC1: _hh_mm,
        0xC2: [_int, DICT_41_ON_OFF],
        # 0xC3: #Specifies ion emission method implemented in humidifier by bit map,
        # Bit 0: negative ion method mounting
        # Bit 1: cluster ion method mounting
        0xC4: [
            _int,
            {
                0x40: "Empty",
                0x41: "Minimum",
                0x42: "Low",
                0x43: "Middium",
                0x44: "High",
                0x45: "Maximum",
            },
        ],
    }

    def __init__(self, host, api_connector=None, instance=0x1):
        self._eojgc = self.class_codes["class_group_code"]
        self._eojcc = self.class_codes["class_code"]
        super().__init__(host, self._eojgc, self._eojcc, instance, api_connector)

    async def getTemperatureSettingValue(self):
        return self.getMessage(0xB3)

    async def setTemperatureSettingValue(self, value):
        self.setMessage(0xB3, value)

    async def getMeasuredTemperature(self):
        return self.getMessage(0xBB)

    async def getAutomaticTemperatureControlSetting(self):
        return self.getMessage(0xB1)

    async def setAutomaticTemperatureControlSetting(self, setting):
        self.setMessage(0xB1, setting)

    async def getOnTimerReservationSetting(self):
        return self.getMessage(0x90)

    async def setOnTimerReservationSetting(self, setting):
        self.setMessage(0x90, setting)

    async def getOnTimerSettingValue(self):
        return self.getMessage(0x91)

    async def setOnTimerSettingValue(self, value):
        self.setMessage(0x91, value)

    async def getOnTimerSettingRelativeTime(self):
        return self.getMessage(0x92)

    async def setOnTimerSettingRelativeTime(self, value):
        self.setMessage(0x92, value)

    async def getOffTimerReservationSetting(self):
        return self.getMessage(0x94)

    async def setOffTimerReservationSetting(self, setting):
        self.setMessage(0x94, setting)

    async def getOffTimerSettingValue(self):
        return self.getMessage(0x95)

    async def setOffTimerSettingValue(self, value):
        self.setMessage(0x95, value)

    async def getOffTimerSettingRelativeTime(self):
        return self.getMessage(0x96)

    async def setOffTimerSettingRelativeTime(self, value):
        self.setMessage(0x96, value)

    async def getExtensionalOperationSetting(self):
        return self.getMessage(0xC0)

    async def setExtensionalOperationSetting(self, setting):
        self.setMessage(0xC0, setting)

    async def getExtensionalOperationTimerTimeSettingValue(self):
        return self.getMessage(0xC1)

    async def setExtensionalOperationTimerTimeSettingValue(self, value):
        self.setMessage(0xC1, value)

    async def getIonEmissionSetting(self):
        return self.getMessage(0xC2)

    async def setIonEmissionSetting(self, setting):
        self.setMessage(0xC2, setting)

    async def getImplementedIonEmissionMethod(self):
        return self.getMessage(0xC3)

    async def getOilAmountLevel(self):
        return self.getMessage(0xC4)



from typing import Any, ByteString
from pychonet.lib.const import MANUFACTURERS

# Data States
DATA_STATE_ON = "on"
DATA_STATE_OFF = "off"
DATA_STATE_LOCK = "locked"
DATA_STATE_UNLOCK = "unlocked"
DATA_STATE_OPEN = "open"
DATA_STATE_CLOSE = "closed"
DATA_STATE_STOP = "stop"

# Like switch type
DICT_41_ON_OFF = {0x41: DATA_STATE_ON, 0x42: DATA_STATE_OFF}
DICT_41_YES_NO = {0x41: "yes", 0x42: "no"}
DICT_41_AUTO_NONAUTO = {0x41: "auto", 0x42: "non-auto"}
DICT_41_UNLOCK_LOCK = {0x42: DATA_STATE_UNLOCK, 0x41: DATA_STATE_LOCK}
DICT_41_OPEN_CLOSED = {0x41: DATA_STATE_OPEN, 0x42: DATA_STATE_CLOSE}
DICT_41_ENABLED_DISABLED = {0x41: "enabled", 0x42: "disabled"}
DICT_41_AVAILABLE_NOT_AVAILABLE = {0x41: "available", 0x42: "not available"}
DICT_41_HEATING_NOT_HEATING = {0x41: "heating", 0x42: "not heating"}
DICT_41_PERMITTED_PROHIBITED = {0x41: "permitted", 0x42: "prohibited"}
DICT_30_TRUE_FALSE = {0x30: True, 0x31: False}
DICT_30_ON_OFF = {0x30: DATA_STATE_ON, 0x31: DATA_STATE_OFF}
DICT_30_OPEN_CLOSED = {0x30: DATA_STATE_OPEN, 0x31: DATA_STATE_CLOSE}

# Like select type
DICT_41_AUTO_8_SPEEDS = {
    0x41: "auto",
    0x31: "minimum",
    0x32: "low",
    0x33: "medium-low",
    0x34: "medium",
    0x35: "medium-high",
    0x36: "high",
    0x37: "very-high",
    0x38: "max",
}


def _swap_dict(d: dict):
    return {v: k for k, v in d.items()}


# ------------ EPC GENERIC FUNCTIONS -------
def _int(edt, values: dict = {}, non_value: Any = "Invalid setting"):  # unsigned int
    data_int = int.from_bytes(edt, "big")
    if len(values):
        return values.get(data_int, non_value)
    else:
        return int.from_bytes(edt, "big")


def _call_int(data: bytes | list):
    if type(data) == list:
        if len(data) == 3:
            return _int(data[0], data[1], data[2])
        elif len(data) == 2:
            return _int(data[0], data[1])
        else:
            return _int(data[0])
    else:
        return _int(data)


def _signed_int(edt):  # signed ints
    return int.from_bytes(edt, byteorder="big", signed=True)


def _hh_mm(edt):  # basic time unit
    hh = int.from_bytes(edt[0:1], "big")
    mm = str(int.from_bytes(edt[1:2], "big")).zfill(2)
    return f"{hh}:{mm}"


def _yyyy_mm_dd(edt):  # basic year unit
    yyyy = str(int.from_bytes(edt[0:4], "big")).zfill(4)
    mm = str(int.from_bytes(edt[4:6], "big")).zfill(2)
    dd = str(int.from_bytes(edt[6:8], "big")).zfill(2)
    return f"{yyyy}:{mm}:{dd}"


def _to_string(edt):
    return edt.decode("utf-8")


def _null_padded_optional_string(edt):
    return (
        edt.decode("utf-8").rstrip("\0") if edt is not None and len(edt) > 0 else None
    )


# Check status of Echonnet Instance
# ----------------- EPC SUPER FUNCTIONS -----------------------------
def _0080(edt):
    return _int(edt, DICT_30_ON_OFF)


def _009X(edt):
    payload = []
    if len(edt) < 17:
        for i in range(1, len(edt)):
            payload.append(edt[i])
        return payload

    for i in range(1, len(edt)):
        code = i - 1
        binary = "{0:08b}".format(edt[i])[::-1]
        for j in range(0, 8):
            if binary[j] == "1":
                EPC = (j + 8) * 0x10 + code
                payload.append(EPC)
    return payload


def _0083(edt, host=None):  # UID
    if edt is not None:
        if len(edt) > 1:
            ops_value = edt[1:].hex()
        else:
            if host is not None:
                digits = host.split(".")
                ops_value = digits[2].zfill(3) + digits[3].zfill(3)
            else:
                ops_value = None
        return ops_value
    return None


def _008A(edt):  # manufacturer
    id = int.from_bytes(edt, "big")
    if id in MANUFACTURERS.keys():
        return MANUFACTURERS[id]
    return id


def _009A(edt):  # cumulative runtime
    if len(edt) > 1:
        value = int.from_bytes(edt[1:], "big")
        time_period_multiplier = 0
        if edt[0] == 0x41:
            time_period_multiplier = 1  # measurement is in seconds
        elif edt[0] == 0x42:
            time_period_multiplier = 60  # measurement is in minutes
        elif edt[0] == 0x43:
            time_period_multiplier = 3600  # measurement is in hours
        elif edt[0] == 0x44:
            time_period_multiplier = 3600 * 24  # measurement is in days
        return value * time_period_multiplier
    return None


EPC_SUPER_FUNCTIONS = {
    0x80: _0080,
    0x83: _0083,
    0x84: _int,
    0x85: _int,
    0x8A: _008A,
    0x8C: _null_padded_optional_string,
    0x9A: _009A,
    0x9D: _009X,
    0x9E: _009X,
    0x9F: _009X,
}



import asyncio
from deprecated import deprecated
from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import (
    DATA_STATE_OFF,
    DATA_STATE_ON,
    DICT_41_ON_OFF,
    DICT_41_OPEN_CLOSED,
    DICT_41_UNLOCK_LOCK,
    _int,
)


# ----- Electric Lock Class -------
@deprecated(reason="Scheduled for removal.")
def _026FEX(edt):
    return _int(edt, DICT_41_UNLOCK_LOCK)


@deprecated(reason="Scheduled for removal.")
def _026FE3(edt):
    return _int(edt, DICT_41_OPEN_CLOSED)


@deprecated(reason="Scheduled for removal.")
def _026FE4(edt):
    return _int(edt, {0x41: "occupant", 0x42: "non-occupant"})


@deprecated(reason="Scheduled for removal.")
def _026FE5(edt):
    return _int(
        edt,
        {
            0x40: "normal",
            0x41: "break_open",
            0x42: "door_open",
            0x43: "maunal_unlocked",
            0x44: "tampered",
        },
    )


@deprecated(reason="Scheduled for removal.")
def _026FE6(edt):
    return _int(edt, DICT_41_ON_OFF)


@deprecated(reason="Scheduled for removal.")
def _026FE7(edt):
    return _int(edt, {0x41: "normal", 0x42: "battery_replacement_notification"})


"""Class for Electric Lock Objects"""


class ElectricLock(EchonetInstance):
    EPC_FUNCTIONS = {
        0xE0: [_int, DICT_41_UNLOCK_LOCK],
        0xE1: [_int, DICT_41_UNLOCK_LOCK],
        0xE2: [_int, DICT_41_UNLOCK_LOCK],
        0xE3: [_int, DICT_41_OPEN_CLOSED],
        0xE4: [_int, {0x41: "occupant", 0x42: "non-occupant"}],
        0xE5: [
            _int,
            {
                0x40: "normal",
                0x41: "break_open",
                0x42: "door_open",
                0x43: "maunal_unlocked",
                0x44: "tampered",
            },
        ],
        0xE6: [_int, DICT_41_ON_OFF],
        0xE7: [_int, {0x41: "normal", 0x42: "battery_replacement_notification"}],
    }

    def __init__(self, host, api_connector=None, instance=0x1):
        self._eojgc = 0x02  # Housing/facility-related device group
        self._eojcc = 0x6F  # Electric Lock
        EchonetInstance.__init__(
            self, host, self._eojgc, self._eojcc, instance, api_connector
        )

    """
    getLockOneStatus get the status of lock one.

    return: A string representing the lock status
    """

    async def getLockOneStatus(self):
        return await self.getMessage(0xE1)



from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import (
    DICT_41_HEATING_NOT_HEATING,
    DICT_41_ON_OFF,
    _int,
    _hh_mm,
)


# 0xB0 - Automatic water heating setting
# 0xB1 - Automatic water temperature control setting
# 0xB2 - Water heating status
# 0xB5 Relative time setting value for manual water heating OFF
# 0xB6 - Tank Operation mode setting
# 0xC0 - Daytime reheating permission setting


# 0xC2 - Alarm Status
def _026BC2(edt):
    alarm_status_bin = int.from_bytes(edt, "big")
    alarm_status = {
        "Out of Hot Water": "Normal",
        "Water leaking": "Normal",
        "Water frozen": "Normal",
    }
    if alarm_status_bin & 0b001:
        alarm_status["Out of Hot Water"] = "Alarm"
    if alarm_status_bin & 0b010:
        alarm_status["Water leaking"] = "Alarm"
    if alarm_status_bin & 0b100:
        alarm_status["Water Frozen"] = "Alarm"
    return alarm_status


# 0xC3 - Hot water supply status
# 0xC4 Relative time setting for keeping bath temperature
# 0xE1 - Measured amount of hot water remaining in tank
# 0xE2 - Tank Capacity
# 0xE3 - Automatic bath water heater mode setting
# 0xE9 - Bathroom Priority setting
# 0xEA - Bath operation status monitor
# 0xE4 - Manual bath reheating function setting
# 0xE5 - Manual bath hot water addition function setting
# 0xE6 - Manual lukewarm water temperature lowering function setting
# 0xE7 - Bath Volume water setting one
# 0xE8 - Bath Volume water setting two
# 0xEE - Bath Volume water setting three
# 0xD4 - Bath Volume water setting four
# 0xD5 - Bath Volume water setting four maximum settable level
# 0x90 - ON timer reservation setting
# 0x91 - ON timer setting
# 0xD6 - Sound Volume Setting
# 0xD7 - Mute Setting
# 0xD8 - Remaining Hot Water volume
# 0xD9 - Surplus electric energy prediction value


def _026BD9(edt):
    starting_MM = str(int.from_bytes(edt[0:1], "big")).zfill(2)
    starting_DD = str(int.from_bytes(edt[1:2], "big")).zfill(2)
    starting_hh_raw = int.from_bytes(edt[2:3], "big")
    starting_hh = str(starting_hh_raw).zfill(2)
    predicted_values_raw = edt[3:]
    predicted_values = {}
    it = iter(predicted_values_raw)
    for x in it:
        chunk = [x, next(it)]
        predicted_values[f"{str(starting_hh_raw).zfill(2)}:00"] = int.from_bytes(
            chunk, "big"
        )
        starting_hh_raw = starting_hh_raw + 1
        if starting_hh_raw == 24:
            starting_hh_raw = 0
    return {
        "start_month": starting_MM,
        "start_day": starting_DD,
        "start_hour": f"{starting_hh}:00",
        "predicted_values": predicted_values,
    }


# 0xDB - Rated power consumptions of H/P unit in winter time
# 0xDC - Rated power consumptions of H/P unit in between seasons
# 0xDD - Rated power consumptions of H/P unit in summer time


# 0xCB - Expected electric energy at daytime heating shift time 1
def _026BCB(edt):
    # 0x00000000-0xFFFFFFFD (0-4,294,967,293 W)
    # 10:00, 13:00, 15:00, 17:00
    # unsigned long x 4 16 byts Wh
    if len(edt) == 16:
        v1 = int.from_bytes(edt[0:4], "big")
        v2 = int.from_bytes(edt[4:8], "big")
        v3 = int.from_bytes(edt[8:12], "big")
        v4 = int.from_bytes(edt[12:16], "big")
    else:
        v1 = v2 = v3 = v4 = None
    return {"10:00": v1, "13:00": v2, "15:00": v3, "17:00": v4}


# 0xCC - Consumption of electric energy per hour 1
def _026BCC(edt):
    # 0x0000-0xFFFD (0-65,533 W)
    # When shifting at 10:00, 13:00, 15:00, and 17:00
    # 0x0000: cleared
    # unsigned short x 4 8 bytes Wh
    if len(edt == 8):
        v1 = int.from_bytes(edt[0:2], "big")
        v2 = int.from_bytes(edt[2:4], "big")
        v3 = int.from_bytes(edt[4:6], "big")
        v4 = int.from_bytes(edt[6:8], "big")
    else:
        v1 = v2 = v3 = v4 = None
    return {"10:00": v1, "13:00": v2, "15:00": v3, "17:00": v4}


# 0xCE - Expected electric energy at daytime heating shift time 2
def _026BCE(edt):
    # 0x00000000-0xFFFFFFFD (0-4,294,967,293 W)
    # 13:00, 15:00, 17:00
    # unsigned long x 3 12 bytes Wh
    if len(edt) == 12:
        v1 = int.from_bytes(edt[0:4], "big")
        v2 = int.from_bytes(edt[4:8], "big")
        v3 = int.from_bytes(edt[8:12], "big")
    else:
        v1 = v2 = v3 = None
    return {"13:00": v1, "15:00": v2, "17:00": v3}


# 0xCF - Consumption of electric energy per hour 2
def _026BCF(edt):
    # 0x0000-0xFFFD (0-65,533W)
    # When shifting at 13:00, 15:00, and 17:00
    # 0x0000: cleared
    # unsigned short x 3 6 bytes Wh
    if len(edt) == 6:
        v1 = int.from_bytes(edt[0:2], "big")
        v2 = int.from_bytes(edt[2:4], "big")
        v3 = int.from_bytes(edt[4:6], "big")
    else:
        v1 = v2 = v3 = None
    return {"13:00": v1, "15:00": v2, "17:00": v3}


class ElectricWaterHeater(EchonetInstance):
    EPC_FUNCTIONS = {
        0xB0: [
            _int,
            {
                0x41: "Automatic",
                0x42: "Manual",
                0x43: "Stop",
            },
        ],
        0xB1: [_int, DICT_41_ON_OFF],
        0xB2: [_int, DICT_41_HEATING_NOT_HEATING],
        0xB3: _int,
        0xB4: _int,
        0xB5: _hh_mm,
        0xB6: [
            _int,
            {
                0x41: "Standard",
                0x42: "Saving",
                0x43: "Extra",
            },
        ],
        0xC0: [
            _int,
            {
                0x41: "Permitted",
                0x42: "Not permitted",
            },
        ],
        0xC1: _int,
        0xC2: _026BC2,
        0xC3: [
            _int,
            {
                0x41: "Supplying hot water",
                0x42: "Stopped",
            },
        ],
        0xC4: _hh_mm,
        0xD1: _int,
        0xD3: _int,
        0xE0: _int,
        0xE1: _int,
        0xE2: _int,
        0xE3: [_int, DICT_41_ON_OFF],
        0xE9: [_int, DICT_41_ON_OFF],
        0xEA: [
            _int,
            {
                0x41: "Filling hot water",
                0x42: "Stopped",
                0x43: "Keeping Temperature",
            },
        ],
        0xE4: [_int, DICT_41_ON_OFF],
        0xE5: [_int, DICT_41_ON_OFF],
        0xE6: [_int, DICT_41_ON_OFF],
        0xE7: _int,
        0xE8: [
            _int,
            {
                0x31: "Level 1",
                0x32: "Level 2",
                0x33: "Level 3",
                0x34: "Level 4",
                0x35: "Level 5",
                0x36: "Level 6",
                0x37: "Level 7",
                0x38: "Level 8",
            },
        ],
        0xEE: _int,
        0xD4: _int,
        0xD5: _int,
        0x90: [_int, DICT_41_ON_OFF],
        0x91: _hh_mm,
        0xD6: _int,
        0xD7: [_int, DICT_41_ON_OFF],
        0xD8: _int,
        0xD9: _026BD9,
        0xDB: _int,
        0xDC: _int,
        0xDD: _int,
        0xC7: [_int, {0x01: "yes", 0x00: "no"}],
        0xC8: [
            _int,
            {
                0x14: "20:00",
                0x15: "21:00",
                0x16: "22:00",
                0x17: "23:00",
                0x18: "24:00",
                0x01: "01:00",
            },
        ],
        0xC9: _int,
        0xCA: [
            _int,
            {
                0x09: "9:00",
                0x0A: "10:00",
                0x0B: "11:00",
                0x0C: "12:00",
                0x0D: "13:00",
                0x0E: "14:00",
                0x0F: "15:00",
                0x10: "16:00",
                0x11: "17:00",
                0x00: "cleared",
            },
        ],
        0xCB: _026BCB,
        0xCC: _026BCC,
        0xCD: [
            _int,
            {
                0x0A: "10:00",
                0x0B: "11:00",
                0x0C: "12:00",
                0x0D: "13:00",
                0x0E: "14:00",
                0x0F: "15:00",
                0x10: "16:00",
                0x11: "17:00",
                0x00: "cleared",
            },
        ],
        0xCE: _026BCE,
        0xCF: _026BCF,
    }

    def __init__(self, host, api_connector=None, instance=0x1):
        self._eojgc = 0x02
        self._eojcc = 0x6B
        EchonetInstance.__init__(
            self, host, self._eojgc, self._eojcc, instance, api_connector
        )



from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import (
    DICT_30_ON_OFF,
    DICT_30_OPEN_CLOSED,
    DICT_41_ON_OFF,
    _hh_mm,
    _int,
)


class RiceCooker(EchonetInstance):
    class_codes = {"class_group_code": 0x03, "class_code": 0xBB}

    EPC_FUNCTIONS = {
        0x80: [_int, DICT_30_ON_OFF],
        0xB0: [_int, DICT_30_OPEN_CLOSED],
        0xB1: [
            _int,
            {
                0x41: "Rice cooking completed",
                0x42: "Rice cooking in progress",
                0x43: "Rice cooking paused",
                0x44: "Rice cooking aborted",
            },
        ],
        0xB2: [_int, {0x41: "Start/Restart", 0x42: "Pause"}],
        0xE1: [_int, DICT_41_ON_OFF],
        0xE5: [_int, {0x41: "Installed", 0x42: "Removed"}],
        0xE6: [_int, {0x41: "Installed", 0x42: "Removed"}],
        0x90: [_int, DICT_41_ON_OFF],
        0x91: _hh_mm,
        0x92: _hh_mm,
    }

    def __init__(self, host, api_connector=None, instance=0x1):
        self._eojgc = self.class_codes["class_group_code"]
        self._eojcc = self.class_codes["class_code"]
        super().__init__(host, self._eojgc, self._eojcc, instance, api_connector)

    async def getOperationStatus(self):
        return await self.getMessage(0x80)

    async def setOperationStatus(self, status):
        return await self.setMessage(0x80, status)

    async def getCoverStatus(self):
        return await self.getMessage(0xB0)

    async def getRiceCookingStatus(self):
        return await self.getMessage(0xB1)

    async def getRiceCookingControlSetting(self):
        return await self.getMessage(0xB2)

    async def setRiceCookingControlSetting(self, setting):
        return await self.setMessage(0xB2, setting)

    async def getWarmerSetting(self):
        return await self.getMessage(0xE1)

    async def setWarmerSetting(self, setting):
        return await self.setMessage(0xE1, setting)

    async def getInnerPotRemovalStatus(self):
        return await self.getMessage(0xE5)

    async def getCoverRemovalStatus(self):
        return await self.getMessage(0xE6)

    async def getRiceCookingReservationSetting(self):
        return await self.getMessage(0x90)

    async def setRiceCookingReservationSetting(self, setting):
        return await self.setMessage(0x90, setting)

    async def getRiceCookingReservationSettingTime(self):
        return await self.getMessage(0x91)

    async def setRiceCookingReservationSettingTime(self, time):
        return await self.setMessage(0x91, time)

    async def getRiceCookingReservationSettingRelativeTime(self):
        return await self.getMessage(0x92)

    async def setRiceCookingReservationSettingRelativeTime(self, time):
        return await self.setMessage(0x92, time)



from deprecated import deprecated
from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import DICT_30_TRUE_FALSE
from pychonet.lib.epc_functions import _int, _swap_dict

ENL_FANSPEED_PERCENT = 0xF0
ENL_FAN_DIRECTION = 0xF1
ENL_FAN_OSCILLATION = 0xF2
ENL_BUZZER = 0xFC
ENL_FAN_POWER = 0x80

DICT_FAN_DIRECTION = {0x41: "forward", 0x42: "reverse"}

FAN_DIRECTION = _swap_dict(DICT_FAN_DIRECTION)

FAN_OSCILLATION = _swap_dict(DICT_30_TRUE_FALSE)


# ----- Ceiling Fan Class -------
# Fan speed in percentage
def _013AF0(edt):
    op_mode = int.from_bytes(edt, "big")
    fan_speed_percentage = int((op_mode - 0x30) * 0x0A)
    return fan_speed_percentage


# Fan Direction
@deprecated(reason="Scheduled for removal.")
def _013AF1(edt):
    return _int(edt, DICT_FAN_DIRECTION)


# Fan Fluctuation
@deprecated(reason="Scheduled for removal.")
def _013AF2(edt):
    return _int(edt, DICT_30_TRUE_FALSE)


# Fan Buzzer
@deprecated(reason="Scheduled for removal.")
def _013AFC(edt):
    return _int(edt, DICT_30_TRUE_FALSE)


"""Class for Celing Fan Objects"""


class CeilingFan(EchonetInstance):
    EPC_FUNCTIONS = {
        0xF0: _013AF0,
        0xF1: [_int, DICT_FAN_DIRECTION],
        0xF2: [_int, DICT_30_TRUE_FALSE],
        0xFC: [_int, DICT_30_TRUE_FALSE],
    }

    def __init__(self, host, api_connector=None, instance=0x1):
        self._eojgc = 0x01  # Air conditioner-related device group
        self._eojcc = 0x3A  # Ceiling Fan
        EchonetInstance.__init__(
            self, host, self._eojgc, self._eojcc, instance, api_connector
        )

    """
    setFanSpeedPercent set the desired fan speed in percentage. Will be rounded to the nearest 10% value. 

    param fans_speed: An int representing the fan speed.
    """

    async def setFanSpeedPercent(self, fan_speed_percent):
        return await self.setMessages(
            [
                {"EPC": ENL_FAN_POWER, "PDC": 0x01, "EDT": 0x30},
                {"EPC": ENL_BUZZER, "PDC": 0x01, "EDT": 0x30},
                {
                    "EPC": ENL_FANSPEED_PERCENT,
                    "PDC": 0x01,
                    "EDT": round(fan_speed_percent / 10) + 0x30,
                },
            ]
        )

    """
    GetFanSpeedPercent gets the current fan speed percentage
    Refer EPC code 0xF0: ('Fan Speed')

    return: A string representing the fan speed
    """

    async def getFanSpeedPercent(self):  # 0xF0
        return await self.getMessage(ENL_FANSPEED_PERCENT)

    """
    setFanDirection set the fan direction 

    param fan_direction: An string representing the fan direction
                            e.g: 'forward', 'reverse'.
    """

    async def setFanDirection(self, fan_direction):
        return await self.setMessages(
            [
                {"EPC": ENL_FAN_POWER, "PDC": 0x01, "EDT": 0x30},
                {"EPC": ENL_BUZZER, "PDC": 0x01, "EDT": 0x30},
                {
                    "EPC": ENL_FAN_DIRECTION,
                    "PDC": 0x01,
                    "EDT": FAN_DIRECTION[fan_direction],
                },
            ]
        )

    """
    getFanDirection gets the current fan direction
    Refer EPC code 0xF1: ('Fan Direction')

    return: A string representing the fan direction
    """

    async def getFanDirection(self):  # 0xF1
        return await self.getMessage(ENL_FAN_DIRECTION)

    """
    setFanOscillation set the fan oscillation

    param fan_oscillation: A boolean representing the fan oscillation status
                            e.g: 'True', 'False'.
    """

    async def setFanOscillation(self, fan_oscillation):
        return await self.setMessages(
            [
                {"EPC": ENL_FAN_POWER, "PDC": 0x01, "EDT": 0x30},
                {"EPC": ENL_BUZZER, "PDC": 0x01, "EDT": 0x30},
                {
                    "EPC": ENL_FAN_OSCILLATION,
                    "PDC": 0x01,
                    "EDT": FAN_OSCILLATION[fan_oscillation],
                },
            ]
        )

    """
    getFanOscillation gets the fan oscillation status
    Refer EPC code 0xF2: ('Fan Oscillation')

    return: A boolean representing the fan oscillation status
    """

    async def getFanOscillation(self):  # 0xF2
        return await self.getMessage(ENL_FAN_OSCILLATION)



from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import _int, _signed_int
import struct

# 0xE2: "Cumulative amount of gas consumption measurement log"
def _0282E2(edt):
      # return array x 48 unsigned long big-endian
      return [x[0] for x in struct.iter_unpack('>L',edt)]


class GasMeter(EchonetInstance):
    EPC_FUNCTIONS = {
        0xE0: _int,  # 0xE0: "Cumulative amount of gas consumption measurement value"
        0xE2: _0282E2  #_0282E2
    }

    def __init__(self, host, api_connector=None, instance=0x1):
        self._eojgc = 0x02
        self._eojcc = 0x82
        EchonetInstance.__init__(
            self, host, self._eojgc, self._eojcc, instance, api_connector
        )



from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import _int, _signed_int


def _0280E2(edt):
    # return array x 48 unsigned long big-endian
    return _int(edt, {0x01: 0.1, 0x02: 0.01})


class ElectricEnergyMeter(EchonetInstance):
    EPC_FUNCTIONS = {
        0xE0: _int,  # 0xE0: "Cumulative amounts of electric energy measurement value",
        0xE2: _0280E2,  # 0xE2: "Cumulative amounts of electric energy unit",
    }

    def __init__(self, host, api_connector=None, instance=0x1):
        self._eojgc = 0x02
        self._eojcc = 0x80
        EchonetInstance.__init__(
            self, host, self._eojgc, self._eojcc, instance, api_connector
        )



import asyncio

from pychonet.lib.const import (
    ENL_GETMAP,
    ENL_MANUFACTURER,
    ENL_PRODUCT_CODE,
    ENL_PORT,
    INSTANCE_LIST,
    ENL_SETMAP,
    ENL_UID,
    GET,
    MESSAGE_TIMEOUT,
    ENL_STATMAP,
    SETRES,
    GETRES,
    INF,
    INFC,
    SETC_SND,
    GET_SNA,
    INF_SNA,
    SETI,
    SETC,
    ENL_MULTICAST_ADDRESS,
)
from pychonet.lib.epc_functions import EPC_SUPER_FUNCTIONS
from pychonet.lib.functions import TIDError, buildEchonetMsg, decodeEchonetMsg


class ECHONETAPIClient:
    def __init__(self, server):
        self._server = server
        self._logger = print
        self._server.subscribe(self.echonetMessageReceived)
        self._state = {}
        self._next_tx_tid = 0x0000
        self._message_list = {}
        self._failure_list = {}
        self._opc_counts = {}
        self._message_timeout = MESSAGE_TIMEOUT
        self._debug_flag = False
        self._update_callbacks = {}
        self._discover_callback = None
        self._waiting = {}

    async def echonetMessageReceived(self, raw_data, addr):
        updated = False
        host = addr[0]
        is_discovery = False

        if self._debug_flag:
            self._logger(
                f"ECHONETLite Message Received from {host} - Raw data is {raw_data}"
            )

        processed_data = decodeEchonetMsg(raw_data)

        if self._debug_flag:
            self._logger(
                f"ECHONETLite Message Received from {host} - Processed data is {processed_data}"
            )

        tid = processed_data["TID"]
        tid_data = self._message_list.get(tid)

        if self._debug_flag:
            self._logger(
                f"ECHONETLite Message Received from {host} - tid_data is {tid_data}"
            )

        isPush = tid_data is None
        seojgc = processed_data["SEOJGC"]
        seojcc = processed_data["SEOJCC"]
        seojci = processed_data["SEOJCI"]
        esv = processed_data["ESV"]

        if self._state.get(host) is None:  # echonet packet arrived we dont know about
            self._logger(f"Unknown ECHONETLite node has been identified - {host}")
            if callable(self._discover_callback):
                if self._debug_flag:
                    self._logger(f"Called _discover_callback('{host}')")
                await self._discover_callback(host)
            return

        key = f"{host}-{seojgc}-{seojcc}-{seojci}"
        esv_set = esv in [SETRES, SETC_SND]
        esv_get = esv in [GETRES, GET_SNA, INF, INF_SNA, INFC]
        if not isPush:
            self._failure_list[tid] = 0
            self._opc_counts[tid] = len(processed_data["OPC"])
        # handle discovery message response
        for opc in processed_data["OPC"]:
            epc = opc["EPC"]
            if seojgc == 0x0E and seojcc == 0xF0:  # Node Profile Class Response
                if epc in [
                    INSTANCE_LIST,
                    ENL_MANUFACTURER,
                    ENL_PRODUCT_CODE,
                    ENL_UID,
                ]:  # process discovery data
                    is_discovery = True
                    await self.process_discovery_data(host, opc)
                else:
                    # @todo handling others
                    """
                    Ex. 0x05: Instance list notification
                    A property to announce the configuration of instances to be disclosed to the network at startup.
                    This property also announces instances held at the self-node each time the configuration of
                    instances disclosed to the network is changed during system operation, such as instance
                    addition or deletion.
                    """
                    """
                    It is desirable to dynamically add and delete entities when there is an increase
                    or decrease in the number of instances within a device already configured with HA,
                    but that will be an issue for the future.
                    """
                    continue
            elif seojgc == 0x0F:  # User definition class group
                if self._debug_flag:
                    self._logger(
                        f"Packet received from {host} for user definition class group 0x0F"
                    )
                    self._logger(f"Full packet details are {processed_data}")
                    self._logger(f"ignoring packet but please notify devs on Github.")
            else:  # process each EPC in order
                if epc == ENL_SETMAP or epc == ENL_GETMAP or epc == ENL_STATMAP:
                    map = EPC_SUPER_FUNCTIONS[epc](opc["EDT"])
                    self._state[host]["instances"][seojgc][seojcc][seojci][epc] = map
                elif epc in (ENL_UID, ENL_MANUFACTURER, ENL_PRODUCT_CODE):
                    try:
                        self._state[host]["instances"][seojgc][seojcc][seojci][
                            epc
                        ] = EPC_SUPER_FUNCTIONS[epc](opc["EDT"])
                    except KeyError as e:
                        if self._debug_flag:
                            self._logger(
                                f"ECHONET _state _key error {e} of {host} this packet contains the following data: {processed_data}"
                            )
                        continue
                        # raise Exception(
                        #     f"ECHONET Packet contains the following data: {processed_data}"
                        # )
                else:
                    if esv_set:
                        if opc["PDC"] > 0 or isPush:
                            if not isPush:
                                self._failure_list[tid] += 1
                            continue
                        if tid_data.get(epc) is None:
                            self._failure_list[tid] += 1
                            if self._debug_flag:
                                self._logger(
                                    f"EDT is not set in send data for EPC '{epc}' - process each EPC"
                                )
                            continue
                        else:
                            # set request data
                            opc["EDT"] = tid_data[epc]
                    elif esv_get:
                        if opc["PDC"] == 0:
                            if not isPush:
                                self._failure_list[tid] += 1
                            continue
                    else:
                        # @todo more esv support
                        continue

                    if (
                        epc
                        not in self._state[host]["instances"][seojgc][seojcc][seojci]
                        or self._state[host]["instances"][seojgc][seojcc][seojci][epc]
                        != opc["EDT"]
                    ):
                        updated = True
                    self._state[host]["instances"][seojgc][seojcc][seojci][epc] = opc[
                        "EDT"
                    ]

        # Markup "discovered"
        if is_discovery and len(self._state[host]["instances"]):
            if self._state[host].get("uid") is None:
                self._state[host]["uid"] = EPC_SUPER_FUNCTIONS[ENL_UID](b"", host)
            if self._state[host].get("manufacturer") is None:
                self._state[host]["manufacturer"] = EPC_SUPER_FUNCTIONS[
                    ENL_MANUFACTURER
                ](0xFFFFFF.to_bytes(3, "big"))
            self._state[host]["discovered"] = True

        # Markup "available"
        _to_available = False
        if host in self._state:
            if not self._state[host].get("available"):
                _to_available = True
                self._state[host]["available"] = True

        if _to_available:
            # Call update callback functions
            # key is f"{host}-{deojgc}-{deojcc}-{deojci}"
            for _key in self._update_callbacks:
                if _key.startswith(host):
                    for update_func in self._update_callbacks[_key]:
                        await update_func(False)
        else:
            # Call update callback functions
            if updated and key in self._update_callbacks:
                for update_func in self._update_callbacks[key]:
                    await update_func(isPush)

        # if we get duplicate packets that have already been processed then dont worry about the message list.
        # but still process them regardless.
        if tid_data is not None:
            del self._message_list[tid]

    async def discover(self, host=ENL_MULTICAST_ADDRESS):
        if host is ENL_MULTICAST_ADDRESS:
            opc = [{"EPC": INSTANCE_LIST}]
        else:
            opc = [
                {"EPC": ENL_MANUFACTURER},
                {"EPC": ENL_PRODUCT_CODE},
                {"EPC": ENL_UID},
                {"EPC": INSTANCE_LIST},
            ]
        return await self.echonetMessage(host, 0x0E, 0xF0, 0x01, GET, opc)

    async def echonetMessage(self, host, deojgc, deojcc, deojci, esv, opc):
        no_res = True if esv is SETI else False
        payload = None
        # Is node profile
        is_node_profile = deojgc == 0x0E and deojcc == 0xF0
        map_epcs = {ENL_STATMAP, ENL_GETMAP, ENL_SETMAP}
        # Check OPC Code
        try:
            if not is_node_profile:
                if esv == GET:
                    check_map = set(
                        self._state[host]["instances"][deojgc][deojcc][deojci].get(
                            ENL_GETMAP, []
                        )
                    )
                    check_map |= map_epcs
                elif esv == SETC:
                    check_map = set(
                        self._state[host]["instances"][deojgc][deojcc][deojci].get(
                            ENL_SETMAP, []
                        )
                    )
                else:
                    check_map = {}
                if len(check_map):
                    checked_opc = []
                    for values in opc:
                        if values.get("EPC") in check_map:
                            checked_opc.append(values)
                    opc = checked_opc
        except KeyError:
            pass

        message_array = {
            "DEOJGC": deojgc,
            "DEOJCC": deojcc,
            "DEOJCI": deojci,
            "ESV": esv,
            "OPC": opc,
        }
        if self._state.get(host) is None:
            self._state[host] = {"instances": {}, "available": True}

        # Consecutive requests to the device must wait for a response
        if self._waiting.get(host) is None:
            self._waiting[host] = 0
        if self._waiting[host] > 0:
            for x in range(0, self._message_timeout):
                # Wait up to 20(0.1*200) seconds depending on the Echonet specifications.
                await asyncio.sleep(0.1)
                if not self._waiting[host]:
                    break
            if self._waiting[host]:
                return False
        self._waiting[host] += 1

        self._next_tx_tid += 1
        tx_tid = self._next_tx_tid
        message_array["TID"] = tx_tid
        try:
            payload = buildEchonetMsg(message_array)
        except TIDError:  # Quashing the rollover bug hopefully once and for all...
            self._next_tx_tid = 1
            tx_tid = self._next_tx_tid
            message_array["TID"] = tx_tid
            payload = buildEchonetMsg(message_array)

        opc_count = len(opc)
        if no_res:
            is_success = True
        else:
            is_success = False
            tid_data = {}
            for opc_data in opc:
                if opc_data.get("EDT") is not None:
                    if isinstance(opc_data["EDT"], int):
                        tid_data[opc_data["EPC"]] = opc_data["EDT"].to_bytes(
                            opc_data["PDC"], "big"
                        )
            self._message_list[tx_tid] = tid_data

        self._server.send(payload, (host, ENL_PORT))

        if not no_res:
            not_timeout = False
            for x in range(0, self._message_timeout):
                # Wait up to 20(0.1*200) seconds depending on the Echonet specifications.
                await asyncio.sleep(0.1)
                # if tx_tid is not in message list then the message listener has received the message
                if self._message_list.get(tx_tid) is None:
                    # Check OPC count in results
                    if not is_node_profile and tx_tid in self._opc_counts:
                        res_opc_count = self._opc_counts[tx_tid]
                        del self._opc_counts[tx_tid]
                        if self._debug_flag:
                            self._logger(
                                f"OPC count in results is {res_opc_count}/{opc_count} from IP {host}."
                            )
                        if res_opc_count < opc_count:
                            self._waiting[host] -= 1
                            raise EchonetMaxOpcError(res_opc_count)

                    # transaction sucessful remove from list
                    if self._failure_list.get(tx_tid, opc_count) < opc_count:
                        is_success = True
                    if tx_tid in self._failure_list:
                        del self._failure_list[tx_tid]
                    not_timeout = True
                    break
            self._waiting[host] -= 1
            if not is_success:
                if self._message_list.get(tx_tid) is not None:
                    del self._message_list[tx_tid]
            if self._state[host]["available"] != not_timeout:
                self._state[host]["available"] = not_timeout
                # Call update callback functions
                # key is f"{host}-{deojgc}-{deojcc}-{deojci}"
                for key in self._update_callbacks:
                    if key.startswith(host):
                        for update_func in self._update_callbacks[key]:
                            await update_func(False)
        else:
            self._waiting[host] -= 1
        return is_success

    async def getAllPropertyMaps(self, host, eojgc, eojcc, eojci):
        return await self.echonetMessage(
            host,
            eojgc,
            eojcc,
            eojci,
            GET,
            [
                {"EPC": ENL_STATMAP},
                {"EPC": ENL_GETMAP},
                {"EPC": ENL_SETMAP},
            ],
        )

    async def getIdentificationInformation(self, host, eojgc, eojcc, eojci):
        return await self.echonetMessage(
            host,
            eojgc,
            eojcc,
            eojci,
            GET,
            [{"EPC": ENL_UID}, {"EPC": ENL_MANUFACTURER}],
        )

    async def process_discovery_data(self, host, opc_data):
        if "discovered" not in self._state[host]:
            if opc_data["EPC"] == ENL_UID:
                self._state[host]["uid"] = EPC_SUPER_FUNCTIONS[ENL_UID](
                    opc_data["EDT"], host
                )
            elif opc_data["EPC"] == ENL_MANUFACTURER:
                self._state[host]["manufacturer"] = EPC_SUPER_FUNCTIONS[
                    ENL_MANUFACTURER
                ](opc_data["EDT"])
            elif opc_data["EPC"] == ENL_PRODUCT_CODE:
                self._state[host]["product_code"] = EPC_SUPER_FUNCTIONS[
                    ENL_PRODUCT_CODE
                ](opc_data["EDT"])
            else:
                edt = bytearray(opc_data["EDT"])
                # 1st byte: Total number of instances
                # 2nd to 253rd bytes: ECHONET object codes (EOJ3 bytes) enumerated
                edtnum = bytearray(edt)[0]
                for x in range(edtnum):
                    eojgc = bytearray(edt)[1 + (3 * x)]
                    eojcc = bytearray(edt)[2 + (3 * x)]
                    eojci = bytearray(edt)[3 + (3 * x)]
                    if eojgc != 0x0F:  # ignore this group code.
                        # populate state table
                        if eojgc not in list(self._state[host]["instances"].keys()):
                            self._state[host]["instances"].update({eojgc: {}})
                        if eojcc not in list(
                            self._state[host]["instances"][eojgc].keys()
                        ):
                            self._state[host]["instances"][eojgc].update({eojcc: {}})
                        if eojci not in list(
                            self._state[host]["instances"][eojgc][eojcc].keys()
                        ):
                            self._state[host]["instances"][eojgc][eojcc][eojci] = {}
                            self._state[host]["instances"][eojgc][eojcc][eojci].update(
                                {ENL_STATMAP: []}
                            )
                            self._state[host]["instances"][eojgc][eojcc][eojci].update(
                                {ENL_SETMAP: []}
                            )
                            self._state[host]["instances"][eojgc][eojcc][eojci].update(
                                {ENL_GETMAP: []}
                            )

    def register_async_update_callbacks(self, host, eojgc, eojcc, eojci, fn):
        key = f"{host}-{eojgc}-{eojcc}-{eojci}"
        if key not in self._update_callbacks:
            self._update_callbacks[key] = []
        self._update_callbacks[key].append(fn)


class EchonetMaxOpcError(Exception):
    pass



from deprecated import deprecated
from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import (
    DICT_41_PERMITTED_PROHIBITED,
    _int,
    _signed_int,
    _hh_mm,
    _to_string,
    _yyyy_mm_dd,
)


@deprecated(reason="Scheduled for removal.")
def _permission_setting(edt):
    return _int(
        edt,
        {
            0x41: "permitted",
            0x42: "Prohibited",
        },
    )


@deprecated(reason="Scheduled for removal.")
def _027DC1(edt):
    return _int(
        edt,
        {
            0x00: "other",
            0x01: "maximum",
            0x02: "surplus",
            0x03: "designatedPower",
            0x04: "designatedCurrent",
        },
    )


@deprecated(reason="Scheduled for removal.")
def _027DC2(edt):
    return _int(
        edt,
        {
            0x00: "other",
            0x01: "maximum",
            0x02: "loadFollowing",
            0x03: "designatedPower",
            0x04: "designatedCurrent",
        },
    )


def _max_min_int(edt):
    max = str(int.from_bytes(edt[0:4], "big"))
    min = str(int.from_bytes(edt[4:8], "big"))
    return max + "/" + min


def _max_min_short_int(edt):
    max = str(int.from_bytes(edt[0:2], "big"))
    min = str(int.from_bytes(edt[2:4], "big"))
    return max + "/" + min


@deprecated(reason="Scheduled for removal.")
def _027DCF(edt):
    return _int(
        edt,
        {
            0x41: "rapidCharging",
            0x42: "charging",
            0x43: "discharging",
            0x44: "standby",
            0x45: "test",
            0x46: "auto",
            0x48: "restart",
            0x49: "capacityRecalculation",
            0x40: "Other",
        },
    )


@deprecated(reason="Scheduled for removal.")
def _027DDB(edt):
    return _int(
        edt,
        {
            0x00: "reversePowerFlowAcceptable",
            0x01: "independent",
            0x02: "reversePowerFlowNotAcceptable",
        },
    )


@deprecated(reason="Scheduled for removal.")
def _027DE6(edt):
    return _int(
        edt,
        {
            0x00: "unknown",
            0x01: "lead",
            0x02: "ni_mh",
            0x03: "ni_cd",
            0x04: "Lithium",
            0x05: "zinc",
            0x06: "alkaline",
        },
    )


class StorageBattery(EchonetInstance):
    DICT_OPERATION_MODE = {
        0x41: "rapidCharging",
        0x42: "charging",
        0x43: "discharging",
        0x44: "standby",
        0x45: "test",
        0x46: "auto",
        0x48: "restart",
        0x49: "capacityRecalculation",
        0x40: "Other",
    }
    EPC_FUNCTIONS = {
        0x83: _to_string,  # Identification number
        0x97: _hh_mm,  # Current time setting
        0x98: _yyyy_mm_dd,  # Current date setting
        0xA0: _int,  # AC effective capacity (charging)
        0xA1: _int,  # AC effective capacity (discharging)
        0xA2: _int,  # AC chargeable capacity
        0xA3: _int,  # AC dischargeable capacity
        0xA4: _int,  # AC chargeable electric energy
        0xA5: _int,  # AC dischargeable electric energy
        0xA6: _int,  # AC charge upper limit setting
        0xA7: _int,  # AC discharge lower limit setting
        0xA8: _int,  # AC measured cumulative charging electric energy
        0xA9: _int,  # AC measured cumulative discharging electric energy"
        0xAA: _int,  # AC charge amount setting value
        0xAB: _int,  # AC discharge amount setting value
        0xC1: [
            _int,
            {
                0x00: "other",
                0x01: "maximum",
                0x02: "surplus",
                0x03: "designatedPower",
                0x04: "designatedCurrent",
            },
        ],  # Charging method
        0xC2: [
            _int,
            {
                0x00: "other",
                0x01: "maximum",
                0x02: "loadFollowing",
                0x03: "designatedPower",
                0x04: "designatedCurrent",
            },
        ],  # Discharging method
        0xC7: _int,  # AC rated electric energy
        0xC8: _max_min_int,  # Minimum/maximum charging electric power
        0xC9: _max_min_int,  # Minimum/maximum discharging electric power
        0xCA: _max_min_short_int,  # Minimum/maximum charging currentt
        0xCB: _max_min_short_int,  # Minimum/maximum discharging current
        0xCC: [
            _int,
            DICT_41_PERMITTED_PROHIBITED,
        ],  # Re-interconnection permission setting
        0xCD: [
            _int,
            DICT_41_PERMITTED_PROHIBITED,
        ],  # "Operation permission setting
        0xCE: [
            _int,
            DICT_41_PERMITTED_PROHIBITED,
        ],  # "Independent operation permission setting
        0xCF: [_int, DICT_OPERATION_MODE],  # Working operation status
        0xD0: _int,  # Rated electric energy
        0xD1: _int,  # Rated capacity
        0xD2: _int,  # Rated voltage
        0xD3: _signed_int,  # Measured instantaneous charging/discharging electric energy
        0xD4: _signed_int,  # Measured instantaneous charging/discharging current
        0xD5: _signed_int,  # Measured instantaneous charging/discharging voltage
        0xD6: _int,  # Measured cumulative discharging electric energy
        # set only 0xD7: "Measured cumulative discharging electric energy” reset setting",
        0xD8: _int,  # Measured cumulative charging electric energy
        # set only 0xD9: "Measured cumulative charging electric energy” reset setting",
        0xDA: [_int, DICT_OPERATION_MODE],  # Operation mode setting
        0xDB: [
            _int,
            {
                0x00: "reversePowerFlowAcceptable",
                0x01: "independent",
                0x02: "reversePowerFlowNotAcceptable",
            },
        ],  # System-interconnected type
        0xDC: _max_min_int,  # Minimum/maxim um charging power (Independent)
        0xDD: _max_min_int,  # Minimum/maxim um discharging power (Independent)
        0xDE: _max_min_short_int,  # Minimum/maxim um charging current (Independent)
        0xDF: _max_min_short_int,  # Minimum/maxim um discharging current (Independent)
        0xE0: _signed_int,  # Charging/discharging amount setting 1
        0xE1: _signed_int,  # Charging/discharging amount setting 2
        0xE2: _int,  # Remaining stored electricity 1
        0xE3: _int,  # Remaining stored electricity 2
        0xE4: _int,  # Remaining stored electricity 3
        0xE5: _int,  # Battery state of health
        0xE6: [
            _int,
            {
                0x00: "unknown",
                0x01: "lead",
                0x02: "ni_mh",
                0x03: "ni_cd",
                0x04: "Lithium",
                0x05: "zinc",
                0x06: "alkaline",
            },
        ],  # Battery type
        0xE7: _int,  # Charging amount setting 1
        0xE8: _int,  # Discharging amount setting 1
        0xE9: _int,  # Charging amount setting 2
        0xEA: _int,  # Discharging amount setting 2
        0xEB: _int,  # Charging electric energy setting
        0xEC: _int,  # Discharging electric energy setting
        0xED: _int,  # Charging current setting
        0xEE: _int,  # Discharging current setting
        0xEF: _int,  # Rated voltage (Independent)
    }

    # WORKING_OPERATION_STATES = {
    #     0x40: "Other",
    #     0x41: "Rapid charging",
    #     0x42: "Charging",
    #     0x43: "Discharging",
    #     0x44: "Standby",
    #     0x45: "Test",
    #     0x46: "Automatic",
    #     0x48: "Restart",
    #     0x49: "Effective capacity recalculation processing",
    # }

    # def _permission_setting(edt):
    #     op_mode = int.from_bytes(edt, "big")
    #     values = {
    #         0x41: "permitted",
    #         0x42: "Prohibited",
    #     }
    #     return values.get(op_mode, "Invalid setting")

    def __init__(self, host, api_connector=None, instance=0x1):
        self._eojgc = 0x02
        self._eojcc = 0x7D
        EchonetInstance.__init__(
            self, host, self._eojgc, self._eojcc, instance, api_connector
        )



from deprecated import deprecated
from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import (
    DATA_STATE_CLOSE,
    DATA_STATE_OPEN,
    DATA_STATE_STOP,
    _int,
)

ENL_OPENSTATE = 0xE0

DICT_41_OPEN_CLOSE_STOP = {
    0x41: DATA_STATE_OPEN,
    0x42: DATA_STATE_CLOSE,
    0x43: DATA_STATE_STOP,
}


@deprecated(reason="Scheduled for removal.")
def _0260EO(edt):
    return _int(edt, DICT_41_OPEN_CLOSE_STOP)


# TODO - complete class definitions
#           0x80: 'Operation status',
#           0x89: 'Fault description (Recoverable faults)',
#           0x90: 'Timer operation setting',
#           0xC2: 'Wind detection status',
#           0xC3: 'Sunlight detection status',
#           0xD0: 'Opening (extension) speed setting',
#           0xD1: 'Closing (retraction) speed setting',
#           0xD2: 'Operation time',
#           0xD4: 'Automatic operation setting',
#           0xE1: 'Degree-of-opening level',
#           0xE2: 'Shade angle setting ',
#           0xE3: 'Open/close (extension/retraction) speed',
#           0xE5: 'Electric lock setting',
#           0xE8: 'Remote operation setting status',
#           0xE9: 'Selective opening (extension) operation setting',
#           0xEA: 'Open/closed (extended/retracted) status',
#           0xEE: 'One-time opening (extension) speed setting',
#           0xEF: 'One-time closing (retraction) speed setting'


"""Class for Electric Blind/Shade Objects"""


class ElectricBlind(EchonetInstance):
    EOJGC = 0x02  # Housing/facility-related device group
    EOJCC = 0x60  # Electrically operated blind/shade

    EPC_FUNCTIONS = {
        0xE0: [_int, DICT_41_OPEN_CLOSE_STOP],
    }

    def __init__(self, host, api_connector=None, instance=0x1):
        EchonetInstance.__init__(
            self, host, self.EOJGC, self.EOJCC, instance, api_connector
        )

    """
    getOpenCloseSetting get the status of the blind.

    return: A string representing the blind/shade state
    """

    def getOpenCloseSetting(self):
        return self.getMessage(ENL_OPENSTATE)



from pychonet.echonetapiclient import ECHONETAPIClient  # noqa
from pychonet.lib.eojx import EOJX_CLASS

from .version import __version__
from .EchonetInstance import EchonetInstance
from .ElectricBlind import ElectricBlind
from .ElectricShutter import ElectricShutter
from .ElectricCurtain import ElectricCurtain
from .ElectricRainSlidingDoor import ElectricRainSlidingDoor
from .ElectricGate import ElectricGate
from .ElectricWindow import ElectricWindow
from .AutomaticEntranceDoor import AutomaticEntranceDoor
from .ElectricLock import ElectricLock
from .ElectricVehicleCharger import ElectricVehicleCharger
from .GeneralLighting import GeneralLighting
from .SingleFunctionLighting import SingleFunctionLighting
from .LightingSystem import LightingSystem
from .HomeAirCleaner import HomeAirCleaner
from .HomeAirConditioner import HomeAirConditioner
from .HomeSolarPower import HomeSolarPower
from .StorageBattery import StorageBattery
from .TemperatureSensor import TemperatureSensor
from .DistributionPanelMeter import DistributionPanelMeter
from .LowVoltageSmartElectricEnergyMeter import LowVoltageSmartElectricEnergyMeter
from .HybridWaterHeater import HybridWaterHeater
from .HotWaterGenerator import HotWaterGenerator
from .FloorHeater import FloorHeater
from .FuelCell import FuelCell
from .GasMeter import GasMeter
from .ElectricEnergyMeter import ElectricEnergyMeter
from .WaterFlowMeter import WaterFlowMeter
from .CeilingFan import CeilingFan
from .ElectricWaterHeater import ElectricWaterHeater


def Factory(host, server, eojgc, eojcc, eojci=0x01):
    instance = None
    if eojgc in EOJX_CLASS:
        if eojcc in EOJX_CLASS[eojgc]:
            instance = f"{eojgc}-{eojcc}"

    """Factory Method"""
    instances = {
        f"{0x00}-{0x11}": TemperatureSensor,
        f"{0x01}-{0x30}": HomeAirConditioner,
        f"{0x01}-{0x35}": HomeAirCleaner,
        f"{0x01}-{0x3A}": CeilingFan,
        f"{0x02}-{0x60}": ElectricBlind,
        f"{0x02}-{0x61}": ElectricShutter,
        f"{0x02}-{0x62}": ElectricCurtain,
        f"{0x02}-{0x63}": ElectricRainSlidingDoor,
        f"{0x02}-{0x64}": ElectricGate,
        f"{0x02}-{0x65}": ElectricWindow,
        f"{0x02}-{0x66}": AutomaticEntranceDoor,
        f"{0x02}-{0x6B}": ElectricWaterHeater,
        f"{0x02}-{0x6F}": ElectricLock,
        f"{0x02}-{0x72}": HotWaterGenerator,
        f"{0x02}-{0x79}": HomeSolarPower,
        f"{0x02}-{0x7B}": FloorHeater,
        f"{0x02}-{0x7C}": FuelCell,
        f"{0x02}-{0x7D}": StorageBattery,
        f"{0x02}-{0x7E}": ElectricVehicleCharger,
        f"{0x02}-{0x80}": ElectricEnergyMeter,
        f"{0x02}-{0x81}": WaterFlowMeter,
        f"{0x02}-{0x82}": GasMeter,
        f"{0x02}-{0x87}": DistributionPanelMeter,
        f"{0x02}-{0x88}": LowVoltageSmartElectricEnergyMeter,
        f"{0x02}-{0x90}": GeneralLighting,
        f"{0x02}-{0x91}": SingleFunctionLighting,
        f"{0x02}-{0xA3}": LightingSystem,
        f"{0x02}-{0xA6}": HybridWaterHeater,
        None: None,
    }
    instance_object = instances.get(instance, None)
    if instance_object is not None:
        return instance_object(host, server, eojci)

    return EchonetInstance(host, eojgc, eojcc, eojci, server)



from deprecated import deprecated
from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import DICT_41_HEATING_NOT_HEATING, _int, _signed_int


# 0xB0 - Automatic water heating setting
@deprecated(reason="Scheduled for removal.")
def _02A6B0(edt):
    return _int(
        edt,
        {
            0x41: "Automatic water heating",
            0x42: "Manual water heating",
            0x43: "Water heating manual stop",
        },
    )


# 0xB2 - Water heating status
# 0xB3 - Heater status
@deprecated(reason="Scheduled for removal.")
def _02A6B2B3(edt):
    return _int(edt, DICT_41_HEATING_NOT_HEATING)


# 0xB6 - Hot water supply mode setting for auxiliary heat source machine
# 0xB7 - Heater mode setting for auxiliary heat source machine
@deprecated(reason="Scheduled for removal.")
def _02A6B6B7(edt):
    return _int(
        edt,
        {
            0x41: "Set",
            0x42: "No setting",
        },
    )


# 0xB8 - Linkage mode setting for solar power generation
@deprecated(reason="Scheduled for removal.")
def _02A6B8(edt):
    return _int(
        edt,
        {
            0x41: "Mode off",
            0x42: "Household consumption",
            0x43: "Prioritizing electricity sales",
            0x44: "Economic efficiency",
        },
    )


# 0xB9 - Solar power generations utilization time
def _02A6B9(edt):
    start_hh = int.from_bytes(edt[0:1], "big")
    start_mm = str(int.from_bytes(edt[1:2], "big")).zfill(2)
    end_hh = int.from_bytes(edt[2:3], "big")
    end_mm = str(int.from_bytes(edt[3:4], "big")).zfill(2)
    return {"start_time": f"{start_hh}:{start_mm}", "end_time": f"{end_hh}:{end_mm}"}


# 0xC3 - Hot water supply status
@deprecated(reason="Scheduled for removal.")
def _02A6C3(edt):
    return _int(
        edt,
        {
            0x41: "Supplying Hot Water",
            0x42: "Stopped",
        },
    )


# 0xE1 - Measured amount of hot water remaining in tank
# 0xE2 - Tank Capacity


class HybridWaterHeater(EchonetInstance):
    EPC_FUNCTIONS = {
        0xB0: [
            _int,
            {
                0x41: "Automatic water heating",
                0x42: "Manual water heating",
                0x43: "Water heating manual stop",
            },
        ],
        0xB2: [_int, DICT_41_HEATING_NOT_HEATING],
        0xB3: [_int, DICT_41_HEATING_NOT_HEATING],
        0xB6: [
            _int,
            {
                0x41: "Set",
                0x42: "No setting",
            },
        ],
        0xB7: [
            _int,
            {
                0x41: "Set",
                0x42: "No setting",
            },
        ],
        0xB8: [
            _int,
            {
                0x41: "Mode off",
                0x42: "Household consumption",
                0x43: "Prioritizing electricity sales",
                0x44: "Economic efficiency",
            },
        ],
        0xB9: _02A6B9,
        0xC3: [
            _int,
            {
                0x41: "Supplying Hot Water",
                0x42: "Stopped",
            },
        ],
        0xE1: _int,
        0xE2: _int,
    }

    def __init__(self, host, api_connector=None, instance=0x1):
        self._eojgc = 0x02
        self._eojcc = 0xA6
        EchonetInstance.__init__(
            self, host, self._eojgc, self._eojcc, instance, api_connector
        )



from deprecated import deprecated
from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import _int, _signed_int


# 0xD0 System-interconnected type
@deprecated(reason="Scheduled for removal.")
def _0279D0(edt):
    return _int(
        edt,
        {
            0x00: "System interconnected (reverse power flow acceptable)",
            0x01: "Independent",
            0x02: "System interconnected (reverse power flow not acceptable)",
            0x03: "Unknown",
        },
    )


# 0xD1 Output power restraint status
@deprecated(reason="Scheduled for removal.")
def _0279D1(edt):
    return _int(
        edt,
        {
            0x41: "Ongoing restraint (output power control)",
            0x42: "Ongoing restraint (except output power control)",
            0x43: "Ongoing restraint (reason for restraint unknown)",
            0x44: "Not restraining",
            0x45: "Unknown",
        },
    )


class HomeSolarPower(EchonetInstance):
    EPC_FUNCTIONS = {
        0xD0: [
            _int,
            {
                0x00: "System interconnected (reverse power flow acceptable)",
                0x01: "Independent",
                0x02: "System interconnected (reverse power flow not acceptable)",
                0x03: "Unknown",
            },
        ],
        0xD1: [
            _int,
            {
                0x41: "Ongoing restraint (output power control)",
                0x42: "Ongoing restraint (except output power control)",
                0x43: "Ongoing restraint (reason for restraint unknown)",
                0x44: "Not restraining",
                0x45: "Unknown",
            },
        ],
        0xE0: _int,
        0xE1: _int,
        0xE2: _int,
        0xE3: _int,
        0xE4: _int,
        0xE5: _int,
        0xE6: _int,
        0xE7: _int,
        0xE8: _int,
        0xE9: _int,
    }

    def __init__(self, host, api_connector, instance=0x1):
        self._eojgc = 0x02
        self._eojcc = 0x79
        EchonetInstance.__init__(
            self, host, self._eojgc, self._eojcc, instance, api_connector
        )

    def getMeasuredInstantPower(self):
        return self.getMessage(0xE0)

    def getMeasuredCumulPower(self):
        return self.getMessage(0xE1)



from deprecated import deprecated
from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import DICT_41_ON_OFF, _int

# ----- Fuel Cell -------


@deprecated(reason="Scheduled for removal.")
def _027CCA(edt):
    return _int(edt, DICT_41_ON_OFF)


@deprecated(reason="Scheduled for removal.")
def _027CCB(edt):
    return _int(
        edt,
        {
            0x41: "Generating",
            0x42: "Stopped",
            0x43: "Starting",
            0x44: "Stopping",
            0x45: "Idling",
        },
    )


@deprecated(reason="Scheduled for removal.")
def _027CD0(edt):
    return _int(
        edt,
        {
            0x01: "System interconnected (reverse power flow acceptable)",
            0x02: "Independent",
            0x03: " System interconnected (reverse power flow not acceptable)",
        },
    )


class FuelCell(EchonetInstance):
    EPC_FUNCTIONS = {
        0xC1: _int,  # Measured temperature of water in water heater
        0xC2: _int,  # Rated power generation output
        0xC3: _int,  # Heating value of hot water storage tank
        0xC4: _int,  # Measured instantaneous power generation output
        0xC5: _int,  # Measured cumulative power generation output
        # set only 0xC6: # Cumulative power generation output reset setting
        0xC7: _int,  # Measured instantaneous gas consumption
        0xC8: _int,  # Measured cumulative gas consumption
        # set only 0xC9: # Cumulative gas consumption reset setting
        0xCA: [_int, DICT_41_ON_OFF],  # Power generation setting
        0xCB: [
            _int,
            {
                0x41: "Generating",
                0x42: "Stopped",
                0x43: "Starting",
                0x44: "Stopping",
                0x45: "Idling",
            },
        ],  # Power generation status
        0xCC: _int,  # Measured in-house instantaneous power consumption
        0xCD: _int,  # Measured in-house cumulative power consumption
        # set only 0xCE: # In-house cumulative power consumption reset
        0xD0: [
            _int,
            {
                0x01: "System interconnected (reverse power flow acceptable)",
                0x02: "Independent",
                0x03: " System interconnected (reverse power flow not acceptable)",
            },
        ],  # System interconnected type
        0xE1: _int,  # Measured remaining hot water amount
        0xE2: _int,  # Tank capacity
    }

    def __init__(self, host, api_connector=None, instance=0x1):
        self._eojgc = 0x02
        self._eojcc = 0x7C
        EchonetInstance.__init__(
            self, host, self._eojgc, self._eojcc, instance, api_connector
        )



from pychonet.EchonetInstance import EchonetInstance
from pychonet.GeneralLighting import ENL_BRIGHTNESS

# ----- Single function lighting class -------


# TODO - implemenet FUNCTIONS
# 	0xB0: 'Illuminance level setting',


"""Class for Single Function Lighting Objects"""


class SingleFunctionLighting(EchonetInstance):
    def __init__(self, host, api_connector=None, instance=0x1):
        self._eojgc = 0x02  # Housing/facility-related device group
        self._eojcc = 0x91  # Single Function Lighting class
        EchonetInstance.__init__(
            self, host, self._eojgc, self._eojcc, instance, api_connector
        )

    """
    getBrightness get the brightness that has been set in the light

    return: A string representing the configured brightness.
    """

    def getBrightness(self):
        return self.getMessage(ENL_BRIGHTNESS)  # ['brightness']

    """
    setBrightness set the temperature of the light

    param temperature: A string representing the desired temperature.
    """

    def setBrightness(self, brightness):
        return self.setMessage(ENL_BRIGHTNESS, int(brightness))



from datetime import datetime
from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import _int, _signed_int

ENL_LVSEEM_COEF = 0xD3
ENL_LVSEEM_DIGITS = 0xD7
ENL_LVSEEM_ENG_UNIT = 0xE1
ENL_LVSEEM_ENG_NOR = 0xE0
ENL_LVSEEM_ENG_REV = 0xE3
ENL_LVSEEM_INSTANT_ENG = 0xE7
ENL_LVSEEM_INSTANT_CUR = 0xE8


def _0288E0(edt):
    if len(edt) < 1:
        return None
    power_val = _int(edt)
    return None if power_val == 4294967294 else power_val  # no measurement


def _0288E1(edt):
    return _int(
        edt,
        {
            0x00: 1,
            0x01: 0.1,
            0x02: 0.01,
            0x03: 0.001,
            0x04: 0.0001,
            0x0A: 10,
            0x0B: 100,
            0x0C: 1000,
            0x0D: 10000,
        },
        None,
    )


def _0288E2(edt):
    """
    1-2 bytes: day for which the
    historical data of measured
    cumulative amounts of electric
    energy is to be retrieved 0x0000–
    0x0063 (0-99)
    3 and succeeding bytes: measured
    cumulative amounts of electric
    energy 0x00000000–0x05F5E0FF
    (0–99,999,999)

    unsigned short + unsigned long × 48
    """
    return "Not implemented"


def _0288E7(edt):
    if len(edt) < 1:
        return None
    power_val = _signed_int(edt)
    as_None = power_val in [
        -2147483648,
        2147483647,
        2147483646,
    ]  # underflow, overflow, no measurement
    return None if as_None else power_val


def _0288E8(edt):
    if len(edt) < 1:
        return {"r_phase_amperes": None, "t_phase_amperes": None}
    r_phase = float(_signed_int(edt[0:2])) / 10  # R Phase
    t_phase = float(_signed_int(edt[2:4])) / 10  # T Phase
    asNone = [-3276.8, 3276.7, 3276.6]  # underflow, overflow, no measurement
    if r_phase in asNone:
        r_phase = None
    if t_phase in asNone:
        t_phase = None
    return {"r_phase_amperes": r_phase, "t_phase_amperes": t_phase}


def _0288EA(edt):
    """
    1–4 bytes: date of measurement
    YYYY: 0x0001–0x270F
    (1–9999)
    MM: 0x01–0x0C (1–12)
    DD: 0x01–0x1F (1–31)
    5–7 bytes: time of measurement
    hh: 0x00–0x17 (0–23)
    mm: 0x00–0x3B (0–59)
    ss: 0x00–0x3B (0–59)
    8–11 bytes: cumulative amounts of
    electric energy measured
    0x00000000–0x05F5E0FF
    (0–99,999,999)

    unsigned short + unsigned char ×2
        +
    unsigned char × 3
        +
    unsigned long
    """
    year = int.from_bytes(edt[0:2], "big")
    month = int.from_bytes(edt[2:3], "big")
    day = int.from_bytes(edt[3:4], "big")
    hour = int.from_bytes(edt[4:5], "big")
    minute = int.from_bytes(edt[5:6], "big")
    second = int.from_bytes(edt[6:7], "big")
    culmative = int.from_bytes(edt[7:], "big")
    return {
        "time": datetime(year, month, day, hour, minute, second).isoformat(),
        "culmative_value": culmative,
    }


def _0288EC(edt):
    """
    1–6 bytes: date and time for which
    the historical data is to be retrieved
    YYYY: 0x0001–0x270F (1–9999)
    MM: 0x01–0x0C (1–12)
    DD: 0x01–0x1F (1–31)
    hh: 0x00–0x17 (0–23)
    mm: 0x00/0x1E (0/30)
    7 byte: number of collection
    segments
    0x01–0x0C (1–12)
    8th and succeeding bytes:
    Measured cumulative amount of
    electric energy (normal direction)
    0x00000000–0x05F5E0FF
    (0–99,999,999)
    Measured cumulative amount of
    electric energy (reverse direction)
    0x00000000–0x05F5E0FF
    (0–99,999,999)

    unsigned short + unsigned char × 4
        +
    unsigned char + (unsigned long + unsigned long) × (Max) 12
    """
    return "Not implemented"


def _0288ED(edt):
    """
    1–6 bytes: date and time for which
    the historical data is to be retrieved
    YYYY: 0x0001–0x270F (1–9999)
    MM: 0x01–0x0C (1–12)
    DD: 0x01–0x1F (1–31)
    hh: 0x00–0x17 (0–23)
    mm: 0x00/0x1E (0/30)
    7 byte: number of collection
    segments
    0x01–0x0C (1–12)

    unsigned short + unsigned char × 4 + unsigned char
    """
    return "Not implemented"


class LowVoltageSmartElectricEnergyMeter(EchonetInstance):
    EPC_FUNCTIONS = {
        0xD3: _int,  # "Coefficient"
        0xD7: _int,  # "Number of effective digits for cumulative amounts of electric energy"
        0xE0: _0288E0,  # "Measured cumulative amount of electric energy (normal direction)"
        0xE1: [
            _int,
            {
                0x00: 1,
                0x01: 0.1,
                0x02: 0.01,
                0x03: 0.001,
                0x04: 0.0001,
                0x0A: 10,
                0x0B: 100,
                0x0C: 1000,
                0x0D: 10000,
            },
            None,
        ],  # "Unit for cumulative amounts of electric energy (normal and reverse directions)"
        # 0xE2: _0288E2, #"Historical data of measured cumulative amounts of electric energy 1 (normal direction)"
        0xE3: _0288E0,  # "Measured cumulative amounts of electric energy (reverse direction)"
        # 0xE4: _0288E2, #"Historical data of measured cumulative amounts of electric energy 1 (reverse direction)"
        # 0xE5: _int, #"Day for which the historical data of measured cumulative amounts of electric energy is to be retrieved 1"
        0xE7: _0288E7,  # "Measured instantaneous electric energy"
        0xE8: _0288E8,  # "Measured instantaneous currents"
        # 0xEA: _0288EA,  # "Cumulative amounts of electric energy measured at fixed time (normal direction)"
        # 0xEB: _0288EA,  # "Cumulative amounts of electric energy measured at fixed time (reverse direction)"
        # 0xEC: _0288EC, #"Historical data of measured cumulative amounts of electric energy 2 (normal and reverse directions)"
        # 0xED: _0288ED, #"Day for which the historical data of measured cumulative amounts of electric energy is to be retrieved 2"
    }

    def __init__(self, host, api_connector, instance=0x1):
        self._eojgc = 0x02
        self._eojcc = 0x88
        EchonetInstance.__init__(
            self, host, self._eojgc, self._eojcc, instance, api_connector
        )



import asyncio
import socket
import struct
from collections import deque

class UDPServer():
    def __init__(self, upload_speed=0, download_speed=0, recv_max_size=256 * 1024, local_ip=None):
        self._upload_speed = upload_speed
        self._download_speed = download_speed
        self._recv_max_size = recv_max_size

        # Get Local IP address
        self._ip = local_ip or [(s.connect(('224.0.23.0', 3610)), s.getsockname()[0], s.close()) for s in [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]][0][1]

        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, 0)
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._sock.setblocking(False)

        # enable multicast
        self._multicast_ips = set()
        self.register_multicast(self._ip)

        self._send_event = asyncio.Event()
        self._send_queue = deque()

        self._subscribers = {}

    # figure out which multicast IP would be used to reach host
    def register_multicast_from_host(self, host):
        # connect to known destination, e.g. via UDP port 80
        temp_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        temp_sock.connect((host, 80))
        with temp_sock:
            local_ip, *_ = temp_sock.getsockname()
            self.register_multicast(local_ip)


    def register_multicast(self, local_ip):
        if (local_ip not in self._multicast_ips):
            self._multicast_ips.add(local_ip)
            self._sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, socket.inet_aton(local_ip))
            mreq = struct.pack("=4s4s", socket.inet_aton("224.0.23.0"), socket.inet_aton(local_ip))
            self._sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)

    # region Interface
    def run(self, host, port, loop):
        self.loop = loop

        self._sock.bind((host, port))


        self._connection_made()

        self._run_future(self._send_periodically(), self._recv_periodically())

    def subscribe(self, fut):
        self._subscribers[id(fut)] = fut

    def unsubscribe(self, fut):
        self._subscribers.pop(id(fut), None)

    def send(self, data, addr):
        if not isinstance(data, (bytes, bytearray)):
            raise TypeError("Param 'data' should be bytes or bytearray type")

        self._send_queue.append((data, addr))
        self._send_event.set()

    # endregion

    def _run_future(self, *args):
        for fut in args:
            asyncio.ensure_future(fut, loop=self.loop)

    def _sock_recv(self, fut=None, registered=False):
        fd = self._sock.fileno()

        if fut is None:
            fut = self.loop.create_future()

        if registered:
            self.loop.remove_reader(fd)

        try:
            data, addr = self._sock.recvfrom(self._recv_max_size)
        except (BlockingIOError, InterruptedError):
            self.loop.add_reader(fd, self._sock_recv, fut, True)
        except Exception as e:
            fut.set_exception(e)
            self._socket_error(e)
        else:
            fut.set_result((data, addr))

        return fut

    def _sock_send(self, data, addr, fut=None, registered=False):
        fd = self._sock.fileno()

        if fut is None:
            fut = self.loop.create_future()

        if registered:
            self.loop.remove_writer(fd)

        if not data:
            return

        try:
            bytes_sent = self._sock.sendto(data, addr)
        except (BlockingIOError, InterruptedError):
            self.loop.add_writer(fd, self._sock_send, data, addr, fut, True)
        except Exception as e:
            fut.set_exception(e)
            self._socket_error(e)
        else:
            fut.set_result(bytes_sent)

        return fut

    async def _throttle(self, data_len, speed=0):
        delay = (data_len / speed) if speed > 0 else 0
        await asyncio.sleep(delay)

    async def _send_periodically(self):
        while True:
            await self._send_event.wait()
            try:
                while self._send_queue:
                    data, addr = self._send_queue.popleft()
                    bytes_sent = await self._sock_send(data, addr)
                    await self._throttle(bytes_sent, self._upload_speed)
            finally:
                self._send_event.clear()

    async def _recv_periodically(self):
        while True:
            data, addr = await self._sock_recv()
            self._notify_subscribers(*self._datagram_received(data, addr))
            await self._throttle(len(data), self._download_speed)

    def _connection_made(self):
        pass

    def _socket_error(self, e):
        pass

    def _datagram_received(self, data, addr):
        return data, addr

    def _notify_subscribers(self, data, addr):
        self._run_future(*(fut(data, addr) for fut in self._subscribers.values()))



from deprecated import deprecated

# TODO fix this ChatGPT created code.
from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import (
    DICT_30_ON_OFF,
    DICT_41_AVAILABLE_NOT_AVAILABLE,
    DICT_41_ENABLED_DISABLED,
    DICT_41_ON_OFF,
    _hh_mm,
    _int,
)


# Static methods for ElectricStorageHeater
@deprecated(reason="Scheduled for removal.")
def _0155B3(payload):
    return _int(payload)


@deprecated(reason="Scheduled for removal.")
def _0155B8(payload):
    return _int(payload)


@deprecated(reason="Scheduled for removal.")
def _0155BB(payload):
    return _int(payload)


@deprecated(reason="Scheduled for removal.")
def _0155BE(payload):
    return _int(payload)


def _0155A0(payload):
    return _int(payload)


@deprecated(reason="Scheduled for removal.")
def _0155A1(payload):
    return _int(payload, DICT_30_ON_OFF)


@deprecated(reason="Scheduled for removal.")
def _0155C0(payload):
    return _int(payload, DICT_41_ON_OFF)


@deprecated(reason="Scheduled for removal.")
def _0155C1(payload):
    return _int(payload)


@deprecated(reason="Scheduled for removal.")
def _0155C2(payload):
    return _int(payload)


@deprecated(reason="Scheduled for removal.")
def _0155C3(payload):
    return _int(payload, DICT_41_ENABLED_DISABLED)


@deprecated(reason="Scheduled for removal.")
def _0155C4(payload):
    return _int(payload, DICT_41_AVAILABLE_NOT_AVAILABLE)


@deprecated(reason="Scheduled for removal.")
def _0155C5(payload):
    return _int(payload)


@deprecated(reason="Scheduled for removal.")
def _0155C6(payload):
    return _hh_mm(payload)


@deprecated(reason="Scheduled for removal.")
def _0155C7(payload):
    return _int(
        payload,
        {
            0x41: "Radiant heat method",
            0x42: "Air heat method",
            0x43: "Radiant and air heat method",
        },
    )


@deprecated(reason="Scheduled for removal.")
def _0155C8(payload):
    return _int(payload, DICT_41_ENABLED_DISABLED)


@deprecated(reason="Scheduled for removal.")
def _0155D0(payload):
    return _int(payload)


@deprecated(reason="Scheduled for removal.")
def _0155D1(payload):
    return _hh_mm(payload)


@deprecated(reason="Scheduled for removal.")
def _0155D2(payload):
    return _hh_mm(payload)


@deprecated(reason="Scheduled for removal.")
def _0155D3(payload):
    return _int(payload)


class ElectricStorageHeater(EchonetInstance):
    class_codes = {"class_group_code": 0x01, "class_code": 0x55}

    EPC_FUNCTIONS = {
        0xB3: _int,
        0xB8: _int,
        0xBB: _int,
        0xBE: _int,
        0xA0: _int,
        0xA1: [_int, DICT_30_ON_OFF],
        0xC0: [_int, DICT_41_ON_OFF],
        0xC1: _int,
        0xC2: _int,
        0xC3: [_int, DICT_41_ENABLED_DISABLED],
        0xC4: [_int, DICT_41_AVAILABLE_NOT_AVAILABLE],
        0xC5: _int,
        0xC6: _hh_mm,
        0xC7: [
            _int,
            {
                0x41: "Radiant heat method",
                0x42: "Air heat method",
                0x43: "Radiant and air heat method",
            },
        ],
        0xC8: [_int, DICT_41_ENABLED_DISABLED],
        0xD0: _int,
        0xD1: _hh_mm,
        0xD2: _hh_mm,
        0xD3: _int,
        0xD4: _hh_mm,
        0xD5: _hh_mm,
    }

    def __init__(self, host, api_connector=None, instance=0x1):
        self._eojgc = self.class_codes["class_group_code"]
        self._eojcc = self.class_codes["class_code"]
        super().__init__(host, self._eojgc, self._eojcc, instance, api_connector)

    async def getTemperatureSetting(self):
        return await self.getMessage(0xB3)

    async def getRatedPowerConsumption(self):
        return await self.getMessage(0xB8)

    async def getMeasuredIndoorTemperature(self):
        return await self.getMessage(0xBB)

    async def getMeasuredOutdoorTemperature(self):
        return await self.getMessage(0xBE)

    async def getAirFlowRateSetting(self):
        return await self.getMessage(0xA0)

    async def setAirFlowRateSetting(self, setting):
        await self.setMessage(0xA0, setting)

    async def getFanOperationStatus(self):
        return await self.getMessage(0xA1)

    async def setFanOperationStatus(self, status):
        await self.setMessage(0xA1, status)

    async def getHeatStorageOperationStatus(self):
        return await self.getMessage(0xC0)

    async def setHeatStorageOperationStatus(self, status):
        await self.setMessage(0xC0, status)

    async def getHeatStorageTemperatureSetting(self):
        return await self.getMessage(0xC1)

    async def setHeatStorageTemperatureSetting(self, temperature):
        await self.setMessage(0xC1, temperature)

    async def getMeasuredStoredHeatTemperature(self):
        return await self.getMessage(0xC2)

    async def getDaytimeHeatStorageSetting(self):
        return await self.getMessage(0xC3)

    async def setDaytimeHeatStorageSetting(self, setting):
        await self.setMessage(0xC3, setting)

    async def getDaytimeHeatStorageAbility(self):
        return await self.getMessage(0xC4)

    async def getMidnightPowerDurationSetting(self):
        return await self.getMessage(0xC5)

    async def setMidnightPowerDurationSetting(self, duration):
        await self.setMessage(0xC5, duration)

    async def getMidnightPowerStartTimeSetting(self):
        return await self.getMessage(0xC6)

    async def setMidnightPowerStartTimeSetting(self, time):
        await self.setMessage(0xC6, time)

    async def getRadiationMethod(self):
        return await self.getMessage(0xC7)

    async def setRadiationMethod(self, method):
        await self.setMessage(0xC7, method)

    async def getChildLockSetting(self):
        return await self.getMessage(0xC8)

    async def setChildLockSetting(self, setting):
        await self.setMessage(0xC8, setting)

    async def getFanTimer1Setting(self):
        return await self.getMessage(0xD0)

    async def setFanTimer1Setting(self, setting):
        await self.setMessage(0xD0, setting)

    async def getFanTimer1ONTimeSetting(self):
        return await self.getMessage(0xD1)

    async def setFanTimer1ONTimeSetting(self, time):
        await self.setMessage(0xD1, time)

    async def getFanTimer1OFFTimeSetting(self):
        return await self.getMessage(0xD2)

    async def setFanTimer1OFFTimeSetting(self, time):
        await self.setMessage(0xD2, time)

    async def getFanTimer2Setting(self):
        return await self.getMessage(0xD3)

    async def setFanTimer2Setting(self, setting):
        await self.setMessage(0xD3, setting)

    async def getFanTimer2ONTimeSetting(self):
        return await self.getMessage(0xD4)

    async def setFanTimer2ONTimeSetting(self, time):
        await self.setMessage(0xD4, time)

    async def getFanTimer2OFFTimeSetting(self):
        return await self.getMessage(0xD5)

    async def setFanTimer2OFFTimeSetting(self, time):
        await self.setMessage(0xD5, time)



from pychonet.EchonetInstance import EchonetInstance
from pychonet.lib.epc_functions import DICT_41_ON_OFF, _int, _hh_mm, _signed_int


class FloorHeater(EchonetInstance):
    EOJGC = 0x02
    EOJCC = 0x7B

    EPC_FUNCTIONS = {
        0xE0: _int,  # "Temperature setting 1",
        0xE1: _int,  # "Temperature setting 2",
        0xD1: _int,  # "Temperature setting 2 – Maximum settable level"
        0xE2: _signed_int,  # "Measured room temperature",
        0xE3: _signed_int,  # "Measured floor temperature",
        # 0xE4: _int,  # "Zone change setting", Sets the target zone for control and gets the number of controllable zones b0–b7 is allocated to 0 to 7 Each bit 1: with control, 0: without control
        0xE5: [
            _int,
            {
                0x41: "Normal",
                0x42: "Modest",
                0x43: "High power",
            },
        ],  # "Special operation setting",
        0xE6: [
            _int,
            {
                0x40: "Timer OFF",
                0x41: "Timer 1",
                0x42: "Timer 2",
            },
        ],  # "Daily timer setting",
        # 0xE7: ,  # "Daily timer setting 1", Set the time in the unit of 30 minutes, dividing 24 hours by 30 minutes and allocated to 6 bytes. Each bit 1: worked 0: stopped
        # 0xE8: ,  # "Daily timer setting 2", Set the time in the unit of 30 minutes, dividing 24 hours by 30 minutes and allocated to 6 bytes. Each bit 1: worked 0: stopped
        0x90: [_int, DICT_41_ON_OFF],  # "ON timer reservation setting",
        0x91: _hh_mm,  # "Time set by ON timer",
        0x92: _hh_mm,  # "Relative ON timer setting",
        0x94: [_int, DICT_41_ON_OFF],  # "OFF timer reservation setting",
        0x95: _hh_mm,  # "Time set by OFF timer",
        0x96: _hh_mm,  # "Relative OFF timer setting",
        0x84: _signed_int,  # "Measured instantaneous power consumption",
        0x85: _signed_int,  # "Measured cumulative electric energy consumption",
        0xE9: _signed_int,  # "Rated power consumption",
        0xEA: [
            _int,
            {
                0x41: "Node unit",
                0x42: "Class unit",
                0x43: "Instance unit",
            },
        ],  # "Power consumption measurement method",
    }

    def __init__(self, host, api_connector=None, instance=0x1):
        self._eojgc = self.EOJGC
        self._eojcc = self.EOJCC
        super().__init__(host, self._eojgc, self._eojcc, instance, api_connector)
